from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL

import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Tells whether the Plone-site needs to be upgraded.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade plone_upgrade_needed --site Plone
[/quote]

""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'plone_upgrade_needed',
        help='Should the Plone site be upgraded?',
        description=DOCS)
    command.set_defaults(func=plone_upgrade_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)


@with_api_requestor
@error_handling
def plone_upgrade_command(args, requestor):
    response = requestor.GET('plone_upgrade_needed')
    print(response.text)
    if response.text.strip() not in ('true', 'false'):
        sys.exit(3)



from ftw.upgrade.jsonapi.exceptions import UnkownAPIAction
from ftw.upgrade.jsonapi.exceptions import WrongAPIVersion
from ftw.upgrade.jsonapi.utils import action
from ftw.upgrade.jsonapi.utils import ErrorHandling
from ftw.upgrade.jsonapi.utils import get_action_discovery_information
from ftw.upgrade.jsonapi.utils import jsonify
from zope.interface import implementer
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces import IPublishTraverse

import re


@implementer(IPublishTraverse)
class APIView(BrowserView):
    api_version = 'v1'

    def __init__(self, *args, **kwargs):
        super(APIView, self).__init__(*args, **kwargs)
        self.requested_api_version = None

    def publishTraverse(self, request, name):
        requested_api_version = None
        if re.match(r'v\d+', name):
            requested_api_version = name

        if requested_api_version == self.api_version:
            return self
        elif requested_api_version is not None:
            with ErrorHandling(self.request.RESPONSE):
                raise WrongAPIVersion(requested_api_version)
            request['TraversalRequestNameStack'] = []
            return ''

        action = getattr(self, name, None)
        if action and getattr(action, 'action_info', None):
            return action

        with ErrorHandling(self.request.RESPONSE):
            raise UnkownAPIAction(name)
        request['TraversalRequestNameStack'] = []
        return ''

    @jsonify
    @action('GET')
    def __call__(self):
        return {'actions': get_action_discovery_information(self),
                'api_version': self.api_version}



from zope.component import queryUtility
from zope.intid.interfaces import IIntIds


def update_intids_after_class_migration(event):
    """Update references to class-migrated objects in the intid utility.

    After a class migration all references to the migrated object need to be
    updated, see:
    https://www.fourdigits.nl/blog/changing-your-packagename#1411464714874959

    Intids KeyReferenceToPersistent instances keep references to the objects
    for which they generate an intid. These references need to be updated by
    telling containing bucket that it has changed. The easiest way to do this
    is to delete and re-add the object to the tree.
    """

    intids = queryUtility(IIntIds)
    # plone.app.intid is in the path but the profile is not installed
    if not intids:
        return

    obj = event.object
    intid = intids.queryId(obj)
    # the object does not have an intid.
    if intid is None:
        return

    reference_to_persistent = intids.refs[intid]
    # refresh buckets by deleting and re-adding objects.
    del intids.refs[intid]
    del intids.ids[reference_to_persistent]
    intids.refs[intid] = reference_to_persistent
    intids.ids[reference_to_persistent] = intid
    reference_to_persistent._p_changed = True



from ftw.upgrade.command.jsonapi import APIRequestor
from ftw.upgrade.command.jsonapi import get_api_url
from ftw.upgrade.command.jsonapi import get_instance_port
from ftw.upgrade.command.jsonapi import get_running_instance
from ftw.upgrade.command.jsonapi import get_zope_url
from ftw.upgrade.command.jsonapi import NoRunningInstanceFound
from ftw.upgrade.command.jsonapi import TempfileAuth
from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from Products.CMFPlone.utils import safe_unicode
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from six.moves import map

import os


class ZopeConfPathStub(object):
    """Stubs a path.py Path object for a zope.conf file.
    """

    def __init__(self, *lines):
        self._text = '\n'.join(map(safe_unicode, lines))

    def text(self):
        return self._text


class TestAPIRequestor(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestAPIRequestor, self).setUp()
        self.write_zconf_with_test_instance()

    def test_GET(self):
        requestor = APIRequestor(HTTPBasicAuth(SITE_OWNER_NAME, TEST_USER_PASSWORD))
        jsondata = requestor.GET('list_plone_sites').json()
        self.assertEqual([{u'id': u'plone',
                           u'path': u'/plone',
                           u'title': u'Plone site'}],
                         jsondata)

    def test_GET_raises_error(self):
        requestor = APIRequestor(HTTPBasicAuth(SITE_OWNER_NAME, TEST_USER_PASSWORD))
        with self.assertRaises(HTTPError) as cm:
            requestor.GET('wrong_action')

        self.assertEqual([u'ERROR',
                          u'Unkown API action',
                          u'There is no API action "wrong_action".'],
            cm.exception.response.json())

    def test_GET_with_params(self):
        requestor = APIRequestor(HTTPBasicAuth(SITE_OWNER_NAME, TEST_USER_PASSWORD),
                                 site='plone')

        requestor.GET('get_profile', site='plone',
                      params={'profileid': 'plone.app.discussion:default'})

    def test_error_when_no_running_instance_found(self):
        self.layer['root_path'].joinpath('parts/instance').rmtree()
        requestor = APIRequestor(HTTPBasicAuth(SITE_OWNER_NAME, TEST_USER_PASSWORD))
        with self.assertRaises(NoRunningInstanceFound):
            requestor.GET('list_plone_sites')

    def test_basic_authentication(self):
        requestor = APIRequestor(HTTPBasicAuth(SITE_OWNER_NAME, TEST_USER_PASSWORD))
        jsondata = requestor.GET('current_user').json()
        self.assertEqual('admin', jsondata)

    def test_tempfile_authentication(self):
        requestor = APIRequestor(TempfileAuth(relative_to=os.getcwd()))
        jsondata = requestor.GET('current_user').json()
        self.assertEqual('system-upgrade', jsondata)


class TestJsonAPIUtils(CommandAndInstanceTestCase):

    def test_get_api_url(self):
        self.write_zconf_with_test_instance()
        test_instance_port = self.layer['port']

        self.assertEqual(
            'http://localhost:{0}/upgrades-api/foo'.format(test_instance_port),
            get_api_url('foo'))

        self.assertEqual(
            'http://localhost:{0}/Plone/upgrades-api/bar'.format(test_instance_port),
            get_api_url('bar', site='Plone'))

        self.assertEqual(
            'http://localhost:{0}/Plone/upgrades-api/baz'.format(test_instance_port),
            get_api_url('baz', site='/Plone/'))

    def test_get_api_url_with_public_url(self):
        self.write_zconf_with_test_instance()
        test_instance_port = self.layer['port']

        os.environ['UPGRADE_PUBLIC_URL'] = 'http://domain.com'
        self.assertEqual(
            'http://localhost:{0}/'
            'VirtualHostBase/http/domain.com:80/mount-point/platform/'
            'VirtualHostRoot/upgrades-api/action'.format(test_instance_port),
            get_api_url('action', site='mount-point/platform'))

        os.environ['UPGRADE_PUBLIC_URL'] = 'https://domain.com'
        self.assertEqual(
            'http://localhost:{0}/'
            'VirtualHostBase/https/domain.com:443/mount-point/platform/'
            'VirtualHostRoot/upgrades-api/action'.format(test_instance_port),
            get_api_url('action', site='mount-point/platform'))

        os.environ['UPGRADE_PUBLIC_URL'] = 'https://domain.com/'
        self.assertEqual(
            'http://localhost:{0}/'
            'VirtualHostBase/https/domain.com:443/mount-point/platform/'
            'VirtualHostRoot/upgrades-api/action'.format(test_instance_port),
            get_api_url('action', site='mount-point/platform'))

        os.environ['UPGRADE_PUBLIC_URL'] = 'https://domain.com/foo'
        self.assertEqual(
            'http://localhost:{0}/'
            'VirtualHostBase/https/domain.com:443/mount-point/platform/'
            'VirtualHostRoot/_vh_foo/upgrades-api/action'.format(test_instance_port),
            get_api_url('action', site='mount-point/platform'))

    def test_get_zope_url_without_zconf(self):
        with self.assertRaises(NoRunningInstanceFound):
            get_zope_url()

    def test_find_first_running_instance_info(self):
        test_instance_port = self.layer['port']
        self.write_zconf('instance1', '1000')
        part2 = self.write_zconf('instance2', test_instance_port)
        self.assertEqual(
            {'port': test_instance_port,
             'path': str(part2)},
            get_running_instance(self.layer['root_path']))

    def test_find_first_running_instance_info_with_network_interface(self):
        test_instance_port = self.layer['port']
        self.write_zconf('instance1', '1000')
        part2 = self.write_zconf('instance2',
                                 '0.0.0.0:{0}'.format(test_instance_port))
        self.assertEqual(
            {'port': test_instance_port,
             'path': str(part2)},
            get_running_instance(self.layer['root_path']))

    def test_find_first_running_instance_info_named_zeoclient(self):
        test_zeoclient_port = self.layer['port']
        self.write_zconf('zeoclient1', '1000')
        part2 = self.write_zconf('zeoclient2', test_zeoclient_port)
        self.assertEqual(
            {'port': test_zeoclient_port,
             'path': str(part2)},
            get_running_instance(self.layer['root_path']))

    def test_get_instance_port_port_only(self):
        zconf = ZopeConfPathStub('<http-server>',
                                 '  address 8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_localhost_interface_prefix(self):
        zconf = ZopeConfPathStub('<http-server>',
                                 '  address 127.0.0.1:8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_public_interface_prefix(self):
        zconf = ZopeConfPathStub('<http-server>',
                                 '  address 127.0.0.1:8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_localhost_ip(self):
        zconf = ZopeConfPathStub('ip-address 127.0.0.1',
                                 '<http-server>',
                                 '  address 127.0.0.1:8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_public_ip(self):
        zconf = ZopeConfPathStub('ip-address 0.0.0.0',
                                 '<http-server>',
                                 '  address 127.0.0.1:8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_no_indent(self):
        zconf = ZopeConfPathStub('<http-server>',
                                 'address 8080',
                                 '</http-server>')
        self.assertEqual(8080, get_instance_port(zconf))

    def test_get_instance_port_not_found(self):
        zconf = ZopeConfPathStub('<http-server>',
                                 '</http-server>')
        self.assertEqual(None, get_instance_port(zconf))



from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import TEST_USER_PASSWORD

import os


class TestUserCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestUserCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_sites_help(self):
        self.upgrade_script('user --help')

    def test_prints_authenticated_user(self):
        exitcode, output = self.upgrade_script('user')
        self.assertEqual(0, exitcode)
        self.assertEqual('Authenticated as "admin".\n', output)

    def test_authentication_by_param(self):
        del os.environ['UPGRADE_AUTHENTICATION']
        exitcode, output = self.upgrade_script('user --auth {0}:{1}'.format(
                SITE_OWNER_NAME, TEST_USER_PASSWORD))
        self.assertEqual(0, exitcode)
        self.assertEqual('Authenticated as "admin".\n', output)

    def test_tempfile_authentication_fallback(self):
        del os.environ['UPGRADE_AUTHENTICATION']
        exitcode, output = self.upgrade_script('user')
        self.assertEqual(0, exitcode)
        self.assertEqual('Authenticated as "system-upgrade".\n', output)

    def test_valid_authentication_format_is_required(self):
        exitcode, output = self.upgrade_script('user --auth=john', assert_exitcode=False)
        self.assertEqual(1, exitcode)
        self.assertEqual('ERROR: Invalid authentication information "john".',
                         output.splitlines()[0])



# pylint: disable=E0211, E0213
# E0211: Method has no argument
# E0213: Method should have "self" as first argument


from zope.component.interfaces import IObjectEvent
from zope.interface import Attribute
from zope.interface import Interface


class IUpgradeLayer(Interface):
    """ftw.upgrade specific browser layer.
    """


class IDuringUpgrade(Interface):
    """Request layer to indicate that upgrades are currently executed."""


class IClassMigratedEvent(IObjectEvent):
    """Fired after the class of an object is migrated.

    This event is fired by an UpgradeStep after its migrate_class method is
    called.
    """


class IUpgradeInformationGatherer(Interface):
    """An adapter adpting the generic setup tool (portal_setup), providing
    a cleaned up list of upgrades grouped by profile.

    - Upgrade groups are flattened for simplicity.
    - Only profiles installed on this plone site are listed.
    - Products.CMFPlone is not listed. I thas its own migration mechanism.
    - Upgrades are grouped by done / proposed.
    """

    def __init__(portal_setup):
        """Adapts portal_setup.
        """

    def get_profiles():
        """Returns upgrades grouped by done / proposed and assigned profile.

        Example output:
        >>> [{'db_version': u'3',
        ...   'product': 'Products.CMFEditions',
        ...   'description': u'Extension profile for default ' + \
        ...       'CMFEditions setup.',
        ...   'title': u'CMFEditions',
        ...   'version': u'3',
        ...   'upgrades': [
        ...       {'haspath': ('3',
        ...                   ),
        ...        'description': None,
        ...        'proposed': False,
        ...        'title': u'Fix portal_historyidhandler',
        ...        'dest': ('3',
        ...                ),
        ...        'ssource': '2.0',
        ...        'sortkey': 0,
        ...        'source': ('2',
        ...                   '0'),
        ...        'done': True,
        ...        'id': '8159946379289711266',
        ...        'sdest': '3'}],
        ...   'path': u'/.../profiles/default',
        ...   'type': 2,
        ...   'id': u'Products.CMFEditions:CMFEditions'}]
        """

    def get_upgrades_by_api_ids(*api_ids):
        """Returns a list of ugprade information dicts for each upgrade which
        is selected with a positional argument.
        The upgrades are ordered in the proposed installation order.
        """


class IExecutioner(Interface):
    """Executes multiple upgrade steps of multiple packages.
    Adapts the generic setup tool (portal_setup).
    """

    def __init__(portal_setup):
        """Adapts portal_setup.
        """

    def install(data):
        """Installs the dict data.
        data example:
        >>> {u'Products.CMFEditions:CMFEditions': '8159946379289711266'}
        """


class IUpgradeStep(Interface):
    """An upgrade step class providing tools and helpers for writing
    upgrade steps.

    Register the classmethod ``upgrade`` in ZCML.
    """

    portal_setup = Attribute('The portal_setup tool.')

    def __call__():
        """This method is implemented in each upgrade step with the
        tasks the upgrade should perform.
        """

    def getToolByName(tool_name):
        """Returns the tool with the name ``tool_name`` of the upgraded
        site.
        """

    def objects(catalog_query, message, logger=None):
        """Queries the catalog (unrestricted) and an iterator with full
        objects.
        The iterator configures and calls a ``ProgressLogger`` with the
        passed ``message``.
        """

    def catalog_rebuild_index(name):
        """Reindex the ``portal_catalog`` index identified by ``name``.
        """

    def catalog_has_index(name):
        """Returns whether there is a catalog index ``name``.
        """

    def catalog_add_index(name, type_, extra=None):
        """Adds a new index to the ``portal_catalog`` tool.
        """

    def catalog_remove_index(name):
        """Removes an index to from ``portal_catalog`` tool.
        """

    def actions_remove_action(category, action_id):
        """Removes an action identified by ``action_id`` from
        the ``portal_actions`` tool from a particulary ``category``.
        """

    def actions_remove_type_action(portal_type, action_id):
        """Removes a ``portal_types`` action from the type identified
        by ``portal_type`` with the action id ``action_id``.
        """

    def set_property(context, key, value, data_type='string'):
        """Set a property with the key ``value`` and the value ``value``
        on the ``context`` safely. The property is created with the
        type ``data_type`` if it does not exist.
        """

    def add_lines_to_property(context, key, lines):
        """Updates a property with key ``key`` on the object ``context``
        adding ``lines``. The property is expected to be of type "lines".
        If the property does not exist it is created.
        """

    def setup_install_profile(profileid, steps=None):
        """Installs the generic setup profile identified by ``profileid``.
        If a list step names is passed with ``steps`` (e.g. ['actions']),
        only those steps are installed. All steps are installed by default.
        """


class IPostUpgrade(Interface):
    """Post upgrade adapters are called after each time upgrades are
    installed using the ``@@ftw.upgrade`` view.

    Using named adapters allows us to have multiple post upgrade adapters.
    The name should be the name of the profile of the package (e.g.
    "ftw.upgrade:default"), so that ftw.upgrade is able to order the adapters
    using the dependency graph.
    By doing this we assure that adapters integration / policy packages are
    executed at the end.

    The adapter adapts the portal and the request.
    """

    def __init__(portal, request):
        """The adapter adapts portal and request.
        """

    def __call__():
        """Runs the post upgrade adapter.
        """


class IUpgradeStepRecorder(Interface):
    """The upgrade step recorder stores which upgrade steps are installed
    per Plone site.

    This makes it possible to track "orphan" upgrade steps, which were merged
    in after installing upgrade steps with a higher version.
    This can happen with long-term branches when timestamps are used
    as versions.
    """

    def __init__(portal, profilename):
        """The IUpgradeStepRecorder is a multiadapter, adapting the Plone site
        and the profile name (string).
        """

    def mark_as_installed(target_version):
        """Marks an upgrade step as installed.
        The upgrade step is identified with the target version.
        """

    def is_installed(target_version):
        """Returns whether an upgrade step was already installed on this
        Plone site (boolean).
        """


class IRecordableHandler(Interface):
    """Marker interface for upgrade step handlers which support recording
    upgrade step installation (see IUpgradeStepRecorder).
    Marking the upgrade step as installed is done by handlers (upgrade step
    methods or classes) which provide this interface.
    this
    """



from __future__ import print_function
from contextlib import closing
from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import TERMINAL

import re
import six
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    Install upgrades or profiles. Either the "--proposed" or the \
    "--upgrades" or the "--profiles" argument must be used.

{t.bold}PROPOSED UPGRADES:{t.normal}
    When using the "--proposed" argument, all proposed upgrade steps are\
 installed, ordered by dependencies and versions.

{t.bold}INSTALL PROPOSED UPGRADES OF SPECIFIC PROFILES:{t.normal}
    The "--proposed" argument optionally accepts a list of profiles, \
for which proposed upgrades are installed. \
When no profile is specified, all proposed upgrades of all profiles \
are installed.

{t.bold}SET OF UPGRADES:{t.normal}
    By using the "--upgrades" argument, one or many upgrades can be \
installed. \
Each upgrade is identified by its API id of format "<dest>@<profileid>".

{t.bold}INSTALL PROFILES:{t.normal}
    By using the "--profiles" argument, one or many profiles can be \
    installed. Each install is identified by profile id.

{t.bold}INSTALL ORDER:{t.normal}
    The upgrades are always reordered before they are installed. \
The profiles are ordered topologically with the GS profile dependency graph. \
Upgrades within each profile are ordered by the source / destination versions.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade install --site Plone --proposed
$ bin/upgrade install --site Plone --proposed --auth admin:admin
$ bin/upgrade install --site Plone --proposed my.package:default other.package:default
$ bin/upgrade install --site Plone --upgrades 3001@my.package:default \
3002@my.package:default
$ bin/upgrade install --site Plone --profiles Products.PloneFormGen:default \
unwanted.addon:uninstall
[/quote]

""".format(t=TERMINAL).strip()


def valid_upgrade_step_id(value):
    if not re.match(r'^\w+@[\w\.]+:\w+$', value):
        raise ValueError(value)
    return value


def valid_profile_id(value):
    if ':' not in value:
        raise ValueError(value)
    return value


def setup_argparser(commands):
    command = commands.add_parser(
        'install',
        help='Install upgrades or profiles.',
        description=DOCS)
    command.set_defaults(func=install_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)

    command.add_argument('--force', '-f', action='store_true',
                         dest='force_reinstall',
                         default=False,
                         help='Force reinstall already installed profiles.')

    group = command.add_mutually_exclusive_group(required=True)
    group.add_argument('--upgrades', '-u', nargs='+',
                       help='One or many upgrade step API ids.',
                       type=valid_upgrade_step_id,
                       metavar='UPGRADE-STEP')
    group.add_argument('--proposed', '-p', nargs='*',
                       help='Installs all proposed upgrades of all or specific profiles.',
                       metavar='PROFILE')
    group.add_argument('--profiles', nargs='+',
                       help='One or many profile ids.',
                       type=valid_profile_id,
                       metavar='PROFILE')

    command.add_argument('--skip-deferrable', '-D',
                         help='Do not propose deferrable upgrades. Only takes '
                              'effect if specified in combination with '
                              '--proposed.',
                         default=False,
                         dest='skip_deferrable',
                         action='store_true')

    command.add_argument('--allow-outdated',
                         help='Allow installing on outdated Plone site. '
                              'By default we do not allow this, '
                              'because it is better to first run the '
                              'plone_upgrade command.',
                         default=False,
                         dest='allow_outdated',
                         action='store_true')


@with_api_requestor
@error_handling
def install_command(args, requestor):
    if args.force_reinstall and not args.profiles:
        print('ERROR: --force can only be used with --profiles.', file=sys.stderr)
        sys.exit(3)

    if args.proposed is not None:
        action = 'execute_proposed_upgrades'
        params = [('profiles:list', name) for name in args.proposed]
        if args.skip_deferrable:
            params.append(('propose_deferrable', False))
    elif args.profiles:
        action = 'execute_profiles'
        params = [('profiles:list', name) for name in set(args.profiles)]
        if args.force_reinstall:
            params.append(('force_reinstall', True))
    else:
        action = 'execute_upgrades'
        params = [('upgrades:list', name) for name in set(args.upgrades)]
    if args.allow_outdated:
        params.append(('allow_outdated', True))

    with closing(requestor.POST(action, params=params,
                                stream=True)) as response:
        for line in response.iter_lines(chunk_size=30):
            line = six.ensure_str(line)

            print(line)

    if line.strip() != 'Result: SUCCESS':
        sys.exit(3)



# Source: http://bugs.python.org/issue12806

from ftw.upgrade.command.terminal import TERMINAL

import argparse
import re
import textwrap


def translate(text):
    if text in ('positional arguments',
                'optional arguments'):
        return TERMINAL.bold(_(text).upper())
    else:
        return _(text)


_ = argparse._
argparse._ = translate


class FlexiFormatter(argparse.RawTextHelpFormatter):
    """FlexiFormatter which respects new line formatting and wraps the rest

    Example:
        >>> parser = argparse.ArgumentParser(formatter_class=FlexiFormatter)
        >>> parser.add_argument('--example', help='''\
        ...     This argument's help text will have this first long line\
        ...     wrapped to fit the target window size so that your text\
        ...     remains flexible.
        ...
        ...         1. This option list
        ...         2. is still persisted
        ...         3. and the option strings get wrapped like this with an\
        ...            indent for readability.
        ...
        ...     You must use backslashes at the end of lines to indicate that\
        ...     you want the text to wrap instead of preserving the newline.
        ...
        ...     As with docstrings, the leading space to the text block is\
        ...     ignored.
        ... ''')
        >>> parser.parse_args(['-h'])
        usage: argparse_formatter.py [-h] [--example EXAMPLE]

        optional arguments:
          -h, --help         show this help message and exit
          --example EXAMPLE  This argument's help text will have this first
                             long line wrapped to fit the target window size
                             so that your text remains flexible.

                                 1. This option list
                                 2. is still persisted
                                 3. and the option strings get wrapped like
                                    this with an indent for readability.

                             You must use backslashes at the end of lines to
                             indicate that you want the text to wrap instead
                             of preserving the newline.

                             As with docstrings, the leading space to the
                             text block is ignored.

    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def _fill_text(self, text, width, indent):
        text = (re.compile('\[quote\].*?\[\/quote\]', re.DOTALL)
                .sub(lambda match: (
                    match.group(0)
                    .replace('\n', '[QUOTE:NEWLINE]')),
                     text))

        text = '\n'.join([indent + line for line
                          in self._split_lines(text, width)])

        text = (re.compile('\[quote\](.*?)\[\/quote\]', re.DOTALL)
                .sub(lambda match: TERMINAL.green(
                    match.group(1)
                    .replace('\n', ' ')
                    .replace('[QUOTE:NEWLINE]',
                             TERMINAL.normal + '\n    ' + TERMINAL.green)
                    .rstrip(' ').strip('\n')),
                     text))
        return text

    def _split_lines(self, text, width):
        lines = list()
        main_indent = TERMINAL.length(re.match(r'( *)', text).group(1))
        # Wrap each line individually to allow for partial formatting
        for line in text.splitlines():

            # Get this line's indent and figure out what indent to use
            # if the line wraps. Account for lists of small variety.
            indent = TERMINAL.length(re.match(r'( *)', line).group(1))
            list_match = re.match(r'( *)(([*-+>]+|\w+\)|\w+\.) +)', line)
            if(list_match):
                sub_indent = indent + TERMINAL.length(list_match.group(2))
            else:
                sub_indent = indent

            # Textwrap will do all the hard work for us
            line = self._whitespace_matcher.sub(' ', line).strip()
            new_lines = textwrap.wrap(
                text=line,
                width=width,
                initial_indent=' ' * (indent - main_indent),
                subsequent_indent=' ' * (sub_indent - main_indent),
            )

            # Blank lines get eaten by textwrap, put it back with [' ']
            lines.extend(new_lines or [' '])

        return lines



from ftw.upgrade.tests.base import CommandAndInstanceTestCase

import transaction


class TestRecookCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestRecookCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('recook --help')

    def test_recook_resources(self):
        with self.assert_resources_recooked():
            exitcode, output = self.upgrade_script('recook -s plone')
            self.assertEqual('OK\n', output)
            transaction.begin()



from Acquisition import aq_inner
from Acquisition import aq_parent
from datetime import datetime
from DateTime import DateTime
from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade import UpgradeStep
from ftw.upgrade.exceptions import NoAssociatedProfileError
from ftw.upgrade.indexing import HAS_INDEXING
from ftw.upgrade.indexing import processQueue
from ftw.upgrade.interfaces import IDuringUpgrade
from ftw.upgrade.interfaces import IUpgradeStep
from ftw.upgrade.tests.base import UpgradeTestCase
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.browserlayer.utils import register_layer
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getFSVersionTuple
from unittest import skipIf
from zope.interface import alsoProvides
from zope.interface import Interface
from zope.interface.verify import verifyClass

import pkg_resources

try:
    from Products.CMFPlone.utils import get_installer
except ImportError:
    get_installer = None

ALLOWED_ROLES_AND_USERS_PERMISSION = 'View'
if getFSVersionTuple() > (5, 2):
    ALLOWED_ROLES_AND_USERS_PERMISSION = 'Access contents information'


class IMyProductLayer(Interface):
    """Dummy class used in test_remove_broken_browserlayer()
    """


class TestUpgradeStep(UpgradeTestCase):

    def setUp(self):
        super(TestUpgradeStep, self).setUp()
        self.portal = self.layer['portal']
        self.portal_setup = getToolByName(self.portal, 'portal_setup')
        self.setup_logging()

        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def test_implements_interface(self):
        verifyClass(IUpgradeStep, UpgradeStep)

    def test_upgrade_classmethod(self):
        with self.assertRaises(NotImplementedError):
            UpgradeStep(self.portal_setup)

    def test_portal_setup_attribute(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(testcase.portal_setup, self.portal_setup)

        Step(self.portal_setup)

    def test_portal_attribute(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(testcase.portal, self.portal)

        Step(self.portal_setup)

    def test_getToolByName(self):
        actions_tool = getToolByName(self.portal, 'portal_actions')
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(
                    actions_tool,
                    self.getToolByName('portal_actions'))

        Step(self.portal_setup)

    def test_objects_method_yields_objects_with_logging(self):
        testcase = self
        create(Builder('folder').titled(u'Foo'))
        create(Builder('folder').titled(u'Bar'))
        create(Builder('folder').titled(u'Baz'))

        object_titles = []

        class Step(UpgradeStep):
            def __call__(self):
                for obj in self.objects({'portal_type': 'Folder'},
                                        'Log message',
                                        logger=testcase.logger,
                                        savepoints=False):
                    object_titles.append(obj.Title())

        Step(self.portal_setup)

        self.assertEqual(set(['Foo', 'Bar', 'Baz']), set(object_titles))

        self.assertEqual(['STARTING Log message',
                          '1 of 3 (33%): Log message',
                          'DONE Log message'],
                         self.get_log())

    def test_objects_modifying_catalog_does_not_reduce_result_set(self):
        old_date = DateTime(2010, 1, 1)
        new_date = DateTime(2012, 3, 3)
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))
        create(Builder('folder').with_modification_date(old_date))

        data = {'processed_folders': 0}

        class Step(UpgradeStep):
            def __call__(self):
                for obj in self.objects({'modified': old_date}, ''):
                    obj.setModificationDate(new_date)
                    obj.reindexObject(idxs=['modified'])
                    data['processed_folders'] += 1

        Step(self.portal_setup)
        self.assertEqual(
            4, data['processed_folders'],
            'Updating catalog reduced result set while iterating over it!!!')

    @skipIf(not HAS_INDEXING,
            'Tests must only run when indexing is available')
    def test_logs_indexing_progress(self):
        folders = [
            create(Builder('folder')),
            create(Builder('folder')),
            create(Builder('folder')),
        ]

        class Step(UpgradeStep):
            def __call__(self):
                for folder in folders:
                    folder.reindexObject()
                # manually trigger processing reindex queue
                processQueue()

        alsoProvides(self.portal.REQUEST, IDuringUpgrade)
        Step(self.portal_setup)

        self.assertEqual(['1 of 3 (33%): Processing indexing queue'],
                         self.get_log())

    def test_catalog_rebuild_index(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                ctool = self.getToolByName('portal_catalog')
                name = 'modified'
                testcase.assertEqual(1, ctool._catalog.getIndex(name).indexSize())
                self.catalog_remove_index(name)

                self.catalog_add_index(name, 'DateIndex')
                testcase.assertEqual(0, ctool._catalog.getIndex(name).indexSize())
                self.catalog_rebuild_index(name)
                testcase.assertEqual(1, ctool._catalog.getIndex(name).indexSize())

        create(Builder('folder')
               .titled(u'Rebuild Index Test Obj'))

        Step(self.portal_setup)

    def test_catalog_reindex_objects(self):
        testcase = self
        create(Builder('folder'))

        class Step(UpgradeStep):
            def __call__(self):
                ctool = self.getToolByName('portal_catalog')
                name = 'modified'
                testcase.assertEqual(1, ctool._catalog.getIndex(name).indexSize())
                self.catalog_remove_index(name)

                self.catalog_add_index(name, 'DateIndex')
                testcase.assertEqual(0, ctool._catalog.getIndex(name).indexSize())

                self.catalog_reindex_objects({'portal_type': 'Folder'})

                self.catalog_rebuild_index(name)
                testcase.assertEqual(1, ctool._catalog.getIndex(name).indexSize())

        Step(self.portal_setup)

    def test_catalog_reindex_objects_keeps_modification_date(self):
        testcase = self
        folder = create(Builder('folder'))

        modification_date = DateTime('2010/12/31')
        folder.setModificationDate(modification_date)
        folder.reindexObject(idxs=['modified'])

        class Step(UpgradeStep):
            def __call__(self):

                brain = self.catalog_unrestricted_search(
                    {'UID': folder.UID()})[0]
                testcase.assertEqual(modification_date, brain.modified)

                self.catalog_reindex_objects({})

                brain = self.catalog_unrestricted_search(
                    {'UID': folder.UID()})[0]
                testcase.assertEqual(modification_date, brain.modified)

        Step(self.portal_setup)

    def test_catalog_has_index(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertTrue(self.catalog_has_index('sortable_title'))
                testcase.assertFalse(self.catalog_has_index('foo'))

        Step(self.portal_setup)

    def test_catalog_add_and_remove_indexes(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.catalog_has_index('foo'))
                self.catalog_add_index('foo', 'FieldIndex')
                testcase.assertTrue(self.catalog_has_index('foo'))
                self.catalog_remove_index('foo')
                testcase.assertFalse(self.catalog_has_index('foo'))

        Step(self.portal_setup)

    def test_catalog_unrestricted_get_object(self):
        testcase = self
        folder = create(Builder('folder'))
        create(Builder('document').within(folder))

        folder_path = '/'.join(folder.getPhysicalPath())

        class Step(UpgradeStep):
            def __call__(self):
                query = {'path': folder_path,
                         'portal_type': 'Document'}
                brains = self.catalog_unrestricted_search(query)

                brain = brains[0]
                testcase.assertEqual(
                    'ImplicitAcquisitionWrapper',
                    type(brain).__name__)

                obj = self.catalog_unrestricted_get_object(brain)
                testcase.assertEqual(
                    'ImplicitAcquisitionWrapper',
                    type(obj).__name__)

        Step(self.portal_setup)

    def test_catalog_unrestricted_get_object_removes_dead_brains(self):
        """From time to time there are brains in the catalog for which the
        object no longer exists.
        This can happen because of bugs such as when a "deleted"-event-subscriber
        indexes the object.

        The ``catalog_unrestricted_get_object``, which is used internally for
        methods such as ``objects``, should therefore take care and remove those
        dead brains and log a warning while returning ``None``.
        """
        testcase = self
        folder = create(Builder('folder'))
        # Delete folder and suppress events so that the catalog does not notice
        # the object is gone, then try to get the object.
        aq_parent(aq_inner(folder))._delObject(folder.getId(),
                                               suppress_events=True)

        class Step(UpgradeStep):

            def __call__(self):
                brains = self.get_folder_brains()
                testcase.assertEqual(1, len(brains))
                brain ,= brains

                testcase.assertIsNone(
                    self.catalog_unrestricted_get_object(brain),
                    'Should return None in order to work well with .objects()')

                testcase.assertIn(
                    "The object of the brain with rid {!r} no longer exists"
                    " at the path '/plone/folder'; removing the brain.".format(
                        brain.getRID()),
                    testcase.get_log())

                testcase.assertEqual(
                    0, len(self.get_folder_brains()),
                    'Brain should have been uncataloged at this point.')

            def get_folder_brains(self):
                return self.catalog_unrestricted_search({'portal_type': 'Folder'})

        Step(self.portal_setup)

    def test_catalog_unrestricted_search(self):
        testcase = self

        folder = create(Builder('folder'))
        create(Builder('document').titled(u'Page One').within(folder))
        create(Builder('document').titled(u'Page Two').within(folder))

        folder_path = '/'.join(folder.getPhysicalPath())

        class Step(UpgradeStep):
            def __call__(self):
                query = {'path': folder_path,
                         'portal_type': 'Document'}
                brains = self.catalog_unrestricted_search(query)

                testcase.assertEqual(2, len(brains))
                testcase.assertEqual(['page-one', 'page-two'],
                                     [brain.id for brain in brains])
                testcase.assertEqual('ImplicitAcquisitionWrapper',
                                     type(brains[0]).__name__)

                objects = self.catalog_unrestricted_search(
                    query, full_objects=True)
                testcase.assertEqual(2, len(objects))
                objects = list(objects)
                testcase.assertEqual(['page-one', 'page-two'],
                                     [obj.id for obj in objects])
                testcase.assertEqual('ImplicitAcquisitionWrapper',
                                     type(objects[0]).__name__)

        Step(self.portal_setup)

    def test_catalog_unrestricted_search_filters_nonexisting_objects(self):
        """From time to time there are brains in the catalog for which the
        object no longer exists.
        This can happen because of bugs such as when a "deleted"-event-subscriber
        indexes the object.

        The ``catalog_unrestricted_get_object`` may return with ``None`` in
        such a situation, so ``catalog_unrestricted_search`` needs to filter those
        when ``full_objects`` is set to ``True``.
        """
        testcase = self
        folder = create(Builder('folder'))
        # Delete folder and suppress events so that the catalog does not notice
        # the object is gone, then try to get the object.
        aq_parent(aq_inner(folder))._delObject(folder.getId(),
                                               suppress_events=True)

        class Step(UpgradeStep):
            def __call__(self):
                query = {'portal_type': 'Folder'}
                testcase.assertEqual(
                    1, len(self.catalog_unrestricted_search(query)))

                # Setting full_objects to true should skip missing objects..
                testcase.assertEqual(
                    (),
                    tuple(self.catalog_unrestricted_search(query, full_objects=True)))

                # .. and should have removed the broken brain.
                testcase.assertEqual(
                    (),
                    tuple(self.catalog_unrestricted_search(query)))

        Step(self.portal_setup)

    def test_actions_remove_action(self):
        atool = getToolByName(self.portal, 'portal_actions')
        self.assertIn('rss', atool.get('document_actions'))
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertTrue(self.actions_remove_action(
                        'document_actions', 'rss'))
                testcase.assertFalse(self.actions_remove_action(
                        'document_actions', 'rss'))

        Step(self.portal_setup)
        self.assertNotIn('rss', atool.get('document_actions'))

    def test_actions_remove_type_action(self):
        testcase = self

        class Step(UpgradeStep):
            def event_has_action(self, actionid):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                for action in fti._actions:
                    if action.id == actionid:
                        return True

                return False

            def __call__(self):
                testcase.assertTrue(self.event_has_action('view'))
                testcase.assertTrue(
                    self.actions_remove_type_action('Event', 'view'))
                testcase.assertFalse(self.event_has_action('view'))

        Step(self.portal_setup)

    def test_actions_add_type_action(self):
        testcase = self

        class Step(UpgradeStep):
            def get_event_action(self, actionid):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                for action in fti._actions:
                    if action.id == actionid:
                        return action

            def get_action_ids(self):
                ttool = self.getToolByName('portal_types')
                fti = ttool.get('Event')

                return [action.id for action in fti._actions]

            def __call__(self):
                testcase.assertEqual(
                    None, self.get_event_action('additional'))
                testcase.assertNotIn(
                    'additional',
                    self.get_action_ids())

                self.actions_add_type_action(
                    'Event', 'history', action_id='additional', title='Additional',
                    action='string:#', permissions=('View', ))

                testcase.assertIn(
                    'additional',
                    self.get_action_ids())

                action = self.get_event_action('additional')
                testcase.assertEqual('Additional', action.title)
                testcase.assertEqual('string:#', action.action.text)

        Step(self.portal_setup)

    def test_set_property(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.portal.hasProperty('foo'))

                self.set_property(self.portal, 'foo', 'bar')
                testcase.assertTrue(self.portal.hasProperty('foo'))
                testcase.assertEqual('bar', self.portal.getProperty('foo'))

                self.set_property(self.portal, 'foo', 'baz')
                testcase.assertEqual('baz', self.portal.getProperty('foo'))

        Step(self.portal_setup)

    def test_add_lines_to_property(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                key = 'foo-lines'

                testcase.assertFalse(self.portal.hasProperty(key))

                self.add_lines_to_property(self.portal, key, 'foo')
                testcase.assertEqual(('foo',),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ['bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, 'baz')
                testcase.assertEqual(('foo', 'bar', 'baz'),
                                     self.portal.getProperty(key))

                self.set_property(self.portal, key, ('foo',))
                testcase.assertEqual(('foo',),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ['bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

                self.add_lines_to_property(self.portal, key, ('baz',))
                testcase.assertEqual(('foo', 'bar', 'baz'),
                                     self.portal.getProperty(key))

                self.portal.manage_delProperties([key])
                self.add_lines_to_property(self.portal, key, ['foo', 'bar'])
                testcase.assertEqual(('foo', 'bar'),
                                     self.portal.getProperty(key))

        Step(self.portal_setup)

    def test_setup_install_profile(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_file('catalog.xml', self.asset('exclude-from-nav-index.xml')))

        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.catalog_has_index('excludeFromNav'))
                self.setup_install_profile('profile-the.package:default')
                testcase.assertTrue(self.catalog_has_index('excludeFromNav'))

                self.catalog_remove_index('excludeFromNav')
                testcase.assertFalse(self.catalog_has_index('excludeFromNav'))
                self.setup_install_profile('profile-the.package:default', ['catalog'])
                testcase.assertTrue(self.catalog_has_index('excludeFromNav'))

        with self.package_created():
            Step(self.portal_setup)

    @skipIf(pkg_resources.get_distribution('Products.GenericSetup').parsed_version
            < pkg_resources.parse_version('1.8'),
            'Old GenericSetup version does not support dependency_strategy option.')
    def test_setup_install_profile_does_not_reinstall_installed_profiles(self):
        # In upgrades we want to avoid accidental reinstalls of dependencies
        # when they are already installed.
        # This does only work with Products.GenericSetup>=1.8

        self.package.with_profile(
            Builder('genericsetup profile')
            .named('foo')
            .with_fs_version('1000')
            .with_file(
                'properties.xml',
                '<site><property name="foo" type="string">Foo</property></site>'))

        self.package.with_profile(
            Builder('genericsetup profile')
            .named('bar')
            .with_fs_version('1000')
            .with_dependencies('the.package:foo')
            .with_file(
                'properties.xml',
                '<site><property name="bar" type="string">Bar</property></site>'))

        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual(None, self.portal.getProperty('bar'))
                testcase.assertEqual(None, self.portal.getProperty('foo'))

                self.setup_install_profile('profile-the.package:foo')
                testcase.assertEqual(None, self.portal.getProperty('bar'))
                testcase.assertEqual('Foo', self.portal.getProperty('foo'))

                self.portal._updateProperty('foo', 'Custom')
                testcase.assertEqual(None, self.portal.getProperty('bar'))
                testcase.assertEqual('Custom', self.portal.getProperty('foo'))

                self.setup_install_profile('profile-the.package:bar')
                testcase.assertEqual('Bar', self.portal.getProperty('bar'))
                testcase.assertEqual(
                    'Custom', self.portal.getProperty('foo'),
                    'Accidental reinstall of dependency my.package:foo'
                    ' has caused the "foo" property to be reset.')

        with self.package_created():
            Step(self.portal_setup)

    def test_ensure_profile_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_fs_version('1111'))

        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertEqual('unknown', self.get_version())
                self.ensure_profile_installed('profile-the.package:default')
                testcase.assertEqual((u'1111',), self.get_version())
                self.set_version((u'1000',))
                testcase.assertEqual((u'1000',), self.get_version())
                self.ensure_profile_installed('profile-the.package:default')
                testcase.assertEqual(
                    (u'1000',), self.get_version(),
                    'Profile should not have been installed again because it'
                    ' was already installed.')

            def get_version(self):
                return self.getToolByName('portal_setup').getLastVersionForProfile(
                    'the.package:default')

            def set_version(self, version):
                return self.getToolByName('portal_setup').setLastVersionForProfile(
                    'the.package:default', version)

        with self.package_created():
            Step(self.portal_setup)

    def test_install_upgrade_profile(self):
        class AddSiteProperty(UpgradeStep):
            def __call__(self):
                assert not self.portal.getProperty('foo'), \
                    'property "foo" should not yet exist'
                self.install_upgrade_profile()
                assert self.portal.getProperty('foo') == 'bar', \
                    'property "foo" was not created or is incorrect.'

        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .calling(AddSiteProperty)
                          .with_file('properties.xml', self.asset('foo-property.xml'))))

        with self.package_created():
            self.install_profile('the.package:default', '0')
            self.install_profile_upgrades('the.package:default')

    def test_install_upgrade_profile_raises_exception_when_no_profile_defined(self):
        class Step(UpgradeStep):
            def __call__(self):
                self.install_upgrade_profile()

        with self.assertRaises(NoAssociatedProfileError):
            Step(self.portal_setup)

    def test_is_product_installed(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertTrue(self.is_product_installed(
                    'CMFPlacefulWorkflow'))
                self.uninstall_product('CMFPlacefulWorkflow')
                testcase.assertFalse(self.is_product_installed(
                    'CMFPlacefulWorkflow'))

        Step(self.portal_setup)

    def test_is_profile_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_fs_version('1000'))
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.is_profile_installed(
                    'profile-the.package:default'))
                testcase.assertFalse(self.is_profile_installed(
                    'the.package:default'))
                self.setup_install_profile('profile-the.package:default')
                testcase.assertTrue(self.is_profile_installed(
                    'profile-the.package:default'))
                testcase.assertTrue(self.is_profile_installed(
                    'the.package:default'))

        with self.package_created():
            Step(self.portal_setup)

    def test_is_profile_installed_respects_quickinstaller_uninstall(self):
        testcase = self

        class Step(UpgradeStep):
            def __call__(self):
                testcase.assertFalse(self.is_profile_installed(
                    'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow'))
                self.setup_install_profile(
                    'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow')
                testcase.assertTrue(self.is_profile_installed(
                    'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow'))
                self.uninstall_product('Products.CMFPlacefulWorkflow')
                testcase.assertFalse(self.is_profile_installed(
                    'Products.CMFPlacefulWorkflow:CMFPlacefulWorkflow'))

    def test_uninstall_product(self):
        if get_installer is not None:
            quickinstaller = get_installer(self.portal, self.portal.REQUEST)
            quickinstaller.install_product('CMFPlacefulWorkflow')
            is_product_installed = quickinstaller.is_product_installed
        else:
            quickinstaller = getToolByName(self.portal, 'portal_quickinstaller')
            quickinstaller.installProduct('CMFPlacefulWorkflow')
            is_product_installed = quickinstaller.isProductInstalled

        class Step(UpgradeStep):
            def __call__(self):
                self.uninstall_product('CMFPlacefulWorkflow')

        self.assertTrue(
            is_product_installed('CMFPlacefulWorkflow'),
            'CMFPlacefulWorkflow should be installed')
        Step(self.portal_setup)
        self.assertFalse(
            is_product_installed('CMFPlacefulWorkflow'),
            'CMFPlacefulWorkflow should not be installed')

    def test_migrate_class(self):
        folder = create(Builder('folder'))
        subfolder = create(Builder('folder').within(folder))

        class FancyFolder(subfolder.__class__):
            pass

        class Step(UpgradeStep):
            def __call__(self):
                self.migrate_class(subfolder, FancyFolder)

        self.assertIn(subfolder.__class__.__name__,
                      ('ATFolder', 'Folder'))
        Step(self.portal_setup)
        self.assertEqual('FancyFolder', subfolder.__class__.__name__)

    def test_migrate_class_also_updates_provided_interfaces_info(self):
        if getFSVersionTuple() > (5, ):
            from plone.app.contenttypes.content import Link
            from plone.app.contenttypes.interfaces import ILink
            from plone.app.contenttypes.interfaces import IDocument
        else:
            from Products.ATContentTypes.content.link import ATLink as Link
            from Products.ATContentTypes.interfaces import IATLink as ILink
            from Products.ATContentTypes.interfaces import IATDocument as IDocument

        obj = create(Builder('document'))
        self.assertTrue(IDocument.providedBy(obj))
        self.assertFalse(ILink.providedBy(obj))

        class Step(UpgradeStep):
            def __call__(self):
                self.migrate_class(obj, Link)

        Step(self.portal_setup)
        self.assertFalse(IDocument.providedBy(obj),
                         'Document interface not removed in migration')
        self.assertTrue(ILink.providedBy(obj),
                        'Link interface not added in migration')

    def test_remove_broken_browserlayer(self):
        from plone.browserlayer.utils import registered_layers
        from plone.browserlayer.interfaces import ILocalBrowserLayerType
        register_layer(IMyProductLayer, 'my.product')

        class Step(UpgradeStep):
            def __call__(self):
                self.remove_broken_browserlayer('my.product',
                                                'IMyProductLayer')
        Step(self.portal_setup)

        # Check that it worked.  Note that this check fails if you have called
        # registered_layers earlier in this test before calling
        # remove_broken_browserlayer: the previous answer is cached.
        self.assertFalse(IMyProductLayer in registered_layers())
        # Check it a bit more low level.
        sm = self.portal.getSiteManager()

        adapters = sm.utilities._adapters
        self.assertFalse('my.product' in adapters[0][ILocalBrowserLayerType])
        subscribers = sm.utilities._subscribers
        layer_subscribers = subscribers[0][ILocalBrowserLayerType]
        iface_name = 'IMyProductLayer'
        self.assertEqual(len([layer for layer in layer_subscribers['']
                              if layer.__name__ == iface_name]), 0)
        self.assertNotIn((ILocalBrowserLayerType, 'my.product'),
                         sm._utility_registrations)

    def test_remove_broken_browserlayer_doesnt_fail_if_layer_missing(self):
        class Step(UpgradeStep):
            def __call__(self):
                self.remove_broken_browserlayer('my.nonexistent.product',
                                                'IMyNonexistentProductLayer')
        Step(self.portal_setup)

    def test_remove_remove_broken_portlet_manager(self):
        from plone.portlets.interfaces import IPortletManager
        from plone.portlets.interfaces import IPortletManagerRenderer
        from plone.portlets.manager import PortletManager
        from zope.browser.interfaces import IBrowserView
        from zope.publisher.interfaces.browser import IBrowserRequest

        sm = self.portal.getSiteManager()
        sm.registerUtility(component=PortletManager(),
                           provided=IPortletManager,
                           name='my.manager')

        self.assertIsNotNone(sm.adapters.lookup(
            [Interface, IBrowserRequest, IBrowserView],
            IPortletManagerRenderer,
            'my.manager'))
        self.assertIsNotNone(sm.queryUtility(
            IPortletManager, name='my.manager'))

        class Step(UpgradeStep):
            def __call__(self):
                self.remove_broken_portlet_manager('my.manager')
        Step(self.portal_setup)

        self.assertIsNone(sm.adapters.lookup(
            [Interface, IBrowserRequest, IBrowserView],
            IPortletManagerRenderer,
            'my.manager'))
        self.assertIsNone(sm.queryUtility(
            IPortletManager, name='my.manager'))


    def test_update_security_removes_roles_unmanaged_by_workflow(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder'))

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

        folder.manage_permission('Modify portal content', roles=[],
                                 acquire=True)
        folder.manage_permission('List folder contents', roles=[],
                                 acquire=False)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_acquired('Modify portal content', folder)
        self.assert_permission_not_acquired('List folder contents', folder)

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder)
        Step(self.portal_setup)

        self.assert_permission_not_acquired('View', folder)
        self.assert_permission_not_acquired('Modify portal content', folder)
        self.assert_permission_acquired('List folder contents', folder)

    def test_update_security_reindexes_security_indexes(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users_for(folder))
        folder.reindexObjectSecurity()

        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder)
        Step(self.portal_setup)

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users_for(folder))

    def test_update_security_without_reindexing_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')
        folder = create(Builder('folder')
                        .in_state('published'))

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users_for(folder))

        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_security(folder, reindex_security=False)
        Step(self.portal_setup)

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users_for(folder))

    def test_update_workflow_security_updates_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder'))
        self.assert_permission_not_acquired('Modify portal content', folder)
        folder.manage_permission('Modify portal content', roles=[],
                                 acquire=True)
        self.assert_permission_acquired('Modify portal content', folder)

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(['plone_workflow'])
        Step(self.portal_setup)

        self.assert_permission_not_acquired('Modify portal content', folder)

    def test_update_workflow_security_reindexes_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='plone_workflow')
        folder = create(Builder('folder').in_state('published'))

        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(
                    ['plone_workflow'], reindex_security=False)
        Step(self.portal_setup)

        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users_for(folder))

        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security(
                    ['plone_workflow'], reindex_security=True)
        Step(self.portal_setup)

        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users_for(folder))

    def test_update_workflow_security_expects_list_of_workflows(self):
        class Step(UpgradeStep):
            def __call__(self):
                self.update_workflow_security('foo')

        with self.assertRaises(ValueError) as cm:
            Step(self.portal_setup)

        self.assertEqual('"workflows" must be a list of workflow names.',
                         str(cm.exception))

    def test_base_profile_and_target_version_are_stored_in_attribute(self):
        result = {}

        class Step(UpgradeStep):
            def __call__(self):
                result['base_profile'] = self.base_profile
                result['target_version'] = self.target_version

        Step(self.portal_setup,
             base_profile='profile-ftw.upgrade:default',
             target_version=1500)

        self.assertEqual(
            {'base_profile': 'profile-ftw.upgrade:default',
             'target_version': 1500},
            result)

    def set_workflow_chain(self, for_type, to_workflow):
        wftool = getToolByName(self.portal, 'portal_workflow')
        wftool.setChainForPortalTypes((for_type,),
                                      (to_workflow,))

    def assert_permission_acquired(self, permission, obj):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertNotIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to be acquired on %s' % (
                permission, str(obj)))

    def assert_permission_not_acquired(self, permission, obj):
        not_acquired_permissions = self.get_not_acquired_permissions_of(obj)

        self.assertIn(
            permission, not_acquired_permissions,
            'Expected permission "%s" to NOT be acquired on %s' % (
                permission, str(obj)))

    def get_not_acquired_permissions_of(self, obj):
        acquired_permissions = [
            item for item in obj.permission_settings()
            if not item.get('acquire')
        ]

        return [item.get('name') for item in acquired_permissions]

    def get_allowed_roles_and_users_for(self, obj):
        processQueue()  # trigger async indexing
        catalog = getToolByName(self.portal, 'portal_catalog')
        path = '/'.join(obj.getPhysicalPath())
        rid = catalog.getrid(path)
        index_data = catalog.getIndexDataForRID(rid)
        return index_data.get('allowedRolesAndUsers')



from ftw.upgrade.tests.base import CommandAndInstanceTestCase
from Products.CMFPlone.utils import getFSVersionTuple
from unittest import skipIf

import transaction


class TestCombineBundlesCommand(CommandAndInstanceTestCase):
    def setUp(self):
        super(TestCombineBundlesCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_help(self):
        self.upgrade_script('combine_bundles --help')

    @skipIf(getFSVersionTuple() < (5,), 'The test only works on Plone 5+.')
    def test_combine_bundles(self):
        with self.assert_bundles_combined():
            exitcode, output = self.upgrade_script('combine_bundles -s plone')
            self.assertEqual('OK\n', output)
            transaction.begin()



from ftw.upgrade.testing import UPGRADE_FUNCTIONAL_TESTING
from ftw.upgrade.utils import SavepointIterator
from unittest import TestCase

import os
import transaction


class TestSavepointIterator(TestCase):
    layer = UPGRADE_FUNCTIONAL_TESTING

    def setUp(self):
        super(TestSavepointIterator, self).setUp()
        self.iterable = [1, 2, 3, 4, 5]
        self.txn = transaction.get()

    def tearDown(self):
        self.txn.abort()
        os.environ.pop('UPGRADE_SAVEPOINT_THRESHOLD', None)

    def test_creates_savepoints(self):
        self.assertEqual(
            0, self.txn._savepoint_index,
            'A new transaction should not have any savepoints yet')

        iterator = SavepointIterator.build(self.iterable, threshold=5)

        # Consume entire iterator
        result = list(iterator)

        self.assertEqual(
            self.iterable, result,
            'Iterator should yield every item of `iterable`')

        self.assertEqual(
            1, self.txn._savepoint_index,
            'One savepoint should have been created')

    def test_doesnt_create_savepoints_with_threshold_0(self):
        self.assertEqual(
            0, self.txn._savepoint_index,
            'A new transaction should not have any savepoints yet')

        iterator = SavepointIterator.build(self.iterable, threshold=0)

        # Consume entire iterator
        result = list(iterator)

        self.assertEqual(
            self.iterable, result,
            'Iterator should yield every item of `iterable`')

        self.assertEqual(
            0, self.txn._savepoint_index,
            'threshold=0 should never create any savepoints')

    def test_instanciating_iterator_with_nonzero_threshold_raises(self):
        with self.assertRaises(ValueError):
            SavepointIterator(self.iterable, threshold=0)

        with self.assertRaises(ValueError):
            SavepointIterator(self.iterable, threshold=None)

    def test_default_threshold_is_1000(self):
        # 1000 is the application default.
        self.assertEqual(1000, SavepointIterator.get_default_threshold())

    def test_configure_default_threshold_with_environ_variable(self):
        os.environ['UPGRADE_SAVEPOINT_THRESHOLD'] = '333'
        self.assertEqual(333, SavepointIterator.get_default_threshold())

    def test_disable_default_threshold_with_environ_variable(self):
        os.environ['UPGRADE_SAVEPOINT_THRESHOLD'] = 'None'
        self.assertIsNone(SavepointIterator.get_default_threshold())
        os.environ['UPGRADE_SAVEPOINT_THRESHOLD'] = 'none'
        self.assertIsNone(SavepointIterator.get_default_threshold())

    def test_invalid_default_thresold_configuration(self):
        os.environ['UPGRADE_SAVEPOINT_THRESHOLD'] = 'foo'
        with self.assertRaises(ValueError) as cm:
            SavepointIterator.get_default_threshold()
        self.assertEqual("Invalid savepoint threshold 'foo'", str(cm.exception))



from ftw.upgrade.tests.base import CommandTestCase

import os


class TestCommand(CommandTestCase):

    def test_console_script_created(self):
        path = self.layer['root_path'].joinpath('bin', 'upgrade')
        self.assertTrue(os.path.exists(path), 'Missing executable %s' % path)
        self.assertTrue(os.access(path, os.X_OK), '%s should be executable' % path)

    def test_help_executable(self):
        self.upgrade_script('--help')

    def test_version_executable(self):
        self.upgrade_script('--version')



from ftw.builder import Builder
from ftw.builder import create
from ftw.upgrade.placefulworkflow import PlacefulWorkflowPolicyActivator
from ftw.upgrade.tests.base import WorkflowTestCase
from ftw.upgrade.workflow import WorkflowChainUpdater
from ftw.upgrade.workflow import WorkflowSecurityUpdater
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getFSVersionTuple

ALLOWED_ROLES_AND_USERS_PERMISSION = 'View'
if getFSVersionTuple() > (5, 2):
    ALLOWED_ROLES_AND_USERS_PERMISSION = 'Access contents information'


class TestWorkflowChainUpdater(WorkflowTestCase):

    def test_changing_workflow_with_mapping(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_folder_workflow')

        private_folder = create(Builder('folder')
                                .titled(u'Private Folder')
                                .in_state('private'))
        published_folder = create(Builder('folder')
                                  .titled(u'Published Folder')
                                  .in_state('internal'))

        objects = [private_folder, published_folder]
        mapping = {
            ('intranet_folder_workflow', 'folder_workflow'): {
                'private': 'visible',
                'internal': 'published'}}

        with WorkflowChainUpdater(objects, mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='folder_workflow')

        self.assertReviewStates({
            private_folder: 'visible',
            published_folder: 'published'})

    def test_object_security_is_updated(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_workflow')

        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('external'))

        mapping = {
            ('intranet_workflow', 'plone_workflow'): {
                'external': 'published'}}

        with WorkflowChainUpdater([container], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='plone_workflow')

        self.assertSecurityIsUpToDate()

    def test_object_security_is_reindexed(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='intranet_workflow')

        container = create(Builder('folder')
                           .titled(u'Container')
                           .in_state('internal'))

        self.assertIn('Member',
                      self.get_allowed_roles_and_users(for_object=container))

        mapping = {
            ('intranet_workflow', 'plone_workflow'): {
                'internal': 'published'}}

        with WorkflowChainUpdater([container], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='plone_workflow')

        self.assertEqual(
            ['Anonymous'],
            self.get_allowed_roles_and_users(for_object=container))

    def test_workflow_history_is_migrated(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.execute_transition(folder, 'publish')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'published'})

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        with WorkflowChainUpdater([folder], mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'internal'}, -1)

    def test_workflow_history_is_not_migrated_when_migration_disabled(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.assert_workflow_history_length(folder, 1)

        mapping = {
            ('simple_publication_workflow', 'folder_workflow'): {
                'published': 'published'}}
        with WorkflowChainUpdater([folder], mapping,
                                  migrate_workflow_history=False):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='folder_workflow')

        self.assert_workflow_history_length(folder, 0)

    def test_state_is_set_correctly_when_wfhistory_migration_is_disabled(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.assert_workflow_history_length(folder, 1)

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        with WorkflowChainUpdater([folder], mapping,
                                  migrate_workflow_history=False):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 0)
        self.assertReviewStates({folder: 'internal'})

    def test_workflow_history_transition_is_updated_when_mapped(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='simple_publication_workflow')

        folder = create(Builder('folder'))
        self.execute_transition(folder, 'publish')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'publish',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'published'})

        mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'published': 'internal'}}
        transition_mapping = {
            ('simple_publication_workflow', 'intranet_folder_workflow'): {
                'publish': 'show_internally'}}

        with WorkflowChainUpdater([folder], mapping,
                                  transition_mapping=transition_mapping):
            self.set_workflow_chain(for_type='Folder',
                                    to_workflow='intranet_folder_workflow')

        self.assert_workflow_history_length(folder, 2)
        self.assert_workflow_history_entry(folder, {
                'action': 'show_internally',
                'actor': 'test_user_1_',
                'comments': '',
                'review_state': 'internal'}, -1)

    def execute_transition(self, context, transition):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        wftool.doActionFor(context, transition)

    def assert_workflow_history_entry(self, context, expected, index=-1):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        history = wftool.getInfoFor(context, 'review_history')
        self.assertDictContainsSubset(
            expected, history[index],
            'Workflow history entry is wrong.\n{0}'.format(history))

    def assert_workflow_history_length(self, context, length):
        wftool = getToolByName(self.layer['portal'], 'portal_workflow')
        history = wftool.getInfoFor(context, 'review_history')
        self.assertEqual(
            length, len(history),
            'Workflow history length is wrong.\n{0}'.format(history))


class TestWorkflowSecurityUpdater(WorkflowTestCase):

    def test_updates_only_objects_with_specified_workflows(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='folder_workflow')
        folder = create(Builder('folder'))
        folder.manage_permission('View', roles=[],
                                 acquire=True)

        self.set_workflow_chain(for_type='Document',
                                to_workflow='simple_publication_workflow')
        document = create(Builder('document'))
        document.manage_permission('View', roles=[],
                                   acquire=True)

        self.assert_permission_acquired('View', folder)
        self.assert_permission_acquired('View', document)

        updater = WorkflowSecurityUpdater()
        updater.update(['folder_workflow'])

        self.assert_permission_not_acquired(
            'View', folder, 'The folder should have been updated but wasnt.')
        self.assert_permission_acquired(
            'View', document,
            'The document should NOT have been updated but it was.')

    def test_updates_disabling_update_security(self):
        self.set_workflow_chain(for_type='Folder',
                                to_workflow='folder_workflow')
        folder = create(Builder('folder'))
        folder.manage_permission(
            ALLOWED_ROLES_AND_USERS_PERMISSION, roles=['Reader'], acquire=False)
        folder.reindexObjectSecurity()
        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users(for_object=folder))

        updater = WorkflowSecurityUpdater()
        updater.update(['folder_workflow'], reindex_security=False)
        self.assertEqual(['Reader'],
                         self.get_allowed_roles_and_users(for_object=folder))

        updater.update(['folder_workflow'], reindex_security=True)
        self.assertEqual(['Anonymous'],
                         self.get_allowed_roles_and_users(for_object=folder))

    def test_respects_placeful_workflows_when_updating(self):
        container = create(Builder('folder'))
        document = create(Builder('document').within(container))

        self.create_placeful_workflow_policy(
            named='local_workflow',
            with_workflows={'Document': 'simple_publication_workflow'})
        activator = PlacefulWorkflowPolicyActivator(container)
        activator.activate_policy(
            'local_workflow',
            review_state_mapping={
                (None, 'simple_publication_workflow'): {
                    None: 'private'}})

        document.manage_permission('View', roles=[],
                                   acquire=True)
        self.assert_permission_acquired('View', document)

        updater = WorkflowSecurityUpdater()
        updater.update(['simple_publication_workflow'])

        self.assert_permission_not_acquired(
            'View', document,
            'The document should have been updated but was not.')



from __future__ import print_function
from binascii import hexlify
from ftw.upgrade.utils import get_tempfile_authentication_directory
from path import Path
from requests.auth import AuthBase
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from six.moves.urllib.parse import urlparse

import cgi
import hashlib
import hmac
import logging
import os
import re
import requests
import six
import socket
import sys
import tempfile


logger = logging.getLogger('ftw.upgrade')


class NoRunningInstanceFound(Exception):
    pass


class APIRequestor(object):

    def __init__(self, auth, site=None):
        self.session = requests.Session()
        self.session.auth = auth
        self.site = site

    def GET(self, action, site=None, **kwargs):
        return self._make_request('GET', action, site=site, **kwargs)

    def POST(self, action, site=None, **kwargs):
        return self._make_request('POST', action, site=site, **kwargs)

    def _make_request(self, method, action, site=None, **kwargs):
        url = get_api_url(action, site=site or self.site)
        response = self.session.request(method.upper(), url, **kwargs)
        response.raise_for_status()
        return response


class TempfileAuth(AuthBase):
    """A requests authenticator which writes a tempfile with a random
    hash to verify that the client and the server is at the same
    machine and with the same user.
    """

    def __init__(self, relative_to=None):
        self.relative_to = relative_to

    def __call__(self, request):
        self._generate_tempfile()
        value = ':'.join((os.path.basename(self.authfile.name), self.authhash))
        request.headers['x-ftw.upgrade-tempfile-auth'] = value
        return request

    def _generate_tempfile(self):
        directory = self._get_temp_directory()
        self.authhash = hmac.new(hexlify(os.urandom(32)),
                                 hexlify(os.urandom(32)),
                                 hashlib.sha256).hexdigest()
        self.authfile = tempfile.NamedTemporaryFile(
            dir=directory)
        self.authfile.write(six.ensure_binary(self.authhash))
        self.authfile.flush()
        # Make sure the file is readable by the group, so that the service
        # user running Zope can read it even when it is not the creator.
        Path(self.authfile.name).chmod(0o640)

    def _get_temp_directory(self):
        relative_to = self.relative_to or sys.argv[0]
        return get_tempfile_authentication_directory(relative_to)


def add_requestor_authentication_argument(argparse_command):
    argparse_command.add_argument(
        '--auth',
        help='Authentication information: "<username>:<password>"')


def add_site_path_argument(argparse_command):
    argparse_command.add_argument(
        '--verbose', '-v', action='store_true',
        help='Verbose logging.')
    group = argparse_command.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--site', '-s',
        help='Path to the Plone site.')
    group.add_argument(
        '--pick-site', '-S', action='store_true',
        help='Automatically pick the Plone site, when there is exactly one.')
    group.add_argument(
        '--last-site', '-L', action='store_true',
        help='Automatically pick the last Plone site ordered by path.')
    group.add_argument(
        '--all-sites', '-A', action='store_true',
        help='Perform the action on all Plone sites.')


def add_json_argument(argparse_command):
    argparse_command.add_argument('--json',
                                  action='store_true',
                                  help='Print result as JSON.')


def with_api_requestor(func):
    def func_wrapper(args):
        default_auth = os.environ.get('UPGRADE_AUTHENTICATION', None)
        auth_value = args.auth or default_auth
        if auth_value:
            if len(auth_value.split(':')) != 2:
                print('ERROR: Invalid authentication information '
                      '"{0}".'.format(auth_value))
                print('A string of form "<username>:<password>" is required.')
                sys.exit(1)
            auth = HTTPBasicAuth(*auth_value.split(':'))
        else:
            auth = TempfileAuth()

        site = get_plone_site_by_args(args, APIRequestor(auth))
        requestor = APIRequestor(auth, site=site)
        return func(args, requestor)
    func_wrapper.__name__ = func.__name__
    func_wrapper.__doc__ = func.__doc__
    return func_wrapper


def error_handling(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except NoRunningInstanceFound:
            print('ERROR: No running Plone instance detected.')
            sys.exit(1)
        except HTTPError as exc:
            try:
                response = exc.response
                mimetype = cgi.parse_header(
                    response.headers.get('Content-Type'))[0]
                if mimetype == 'application/json':
                    print(': '.join(response.json()))
                    print('>', exc.request.url)
                    print('<', response.status_code, response.reason)
                    sys.exit(1)

                raise

            except Exception as subexc:
                print('Exception while rendering error:', subexc)
                raise exc

    func_wrapper.__name__ = func.__name__
    func_wrapper.__doc__ = func.__doc__
    return func_wrapper


def get_api_url(action, site=None):
    url = get_zope_url()
    public_url = os.environ.get('UPGRADE_PUBLIC_URL', None)
    if public_url:
        url = extend_url_with_virtualhost_config(url, public_url, site)
    elif site:
        url += site.rstrip('/').strip('/') + '/'
    url += 'upgrades-api/'
    url += action
    return url


def extend_url_with_virtualhost_config(zope_url, public_url, site):
    urlinfo = urlparse(public_url)
    # a port is required for the virtual host monster to work nicely.
    if not urlinfo.port:
        ports = {'http': 80, 'https': 443}
        urlinfo = urlinfo._replace(netloc='{0}:{1}'.format(
            urlinfo.hostname,
            ports[urlinfo.scheme]))

    url = zope_url.rstrip('/')
    url += '/VirtualHostBase'
    url += '/' + urlinfo.scheme
    url += '/' + urlinfo.netloc
    if site:
        url += '/' + site.strip('/')
    url += '/VirtualHostRoot'
    for name in urlinfo.path.split('/'):
        if not name:
            continue
        url += '/_vh_' + name
    url += '/'
    return url


def get_zope_url():
    instance = get_running_instance(Path.getcwd())
    if not instance:
        raise NoRunningInstanceFound()
    return 'http://localhost:{0}/'.format(instance['port'])


def get_running_instance(buildout_path):
    for zconf in find_instance_zconfs(buildout_path):
        port = get_instance_port(zconf)
        if not port:
            continue
        if is_port_open(port):
            return {'port': port,
                    'path': zconf.dirname().dirname()}
    return None


def find_instance_zconfs(buildout_path):
    return sorted(
        buildout_path.glob('parts/*/etc/zope.conf')
        + buildout_path.glob('parts/*/etc/wsgi.ini')
    )


def get_instance_port(zconf):
    # zope.conf
    match = re.search(r'\saddress ([\d.]*:)?(\d+)', zconf.text())
    if match:
        return int(match.group(2))
    # wsgi.ini
    match = re.search(r'\slisten = ([\d.]*:)?(\d+)', zconf.text())
    if match:
        return int(match.group(2))
    return None


def is_port_open(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex(('127.0.0.1', port))
    sock.close()
    return result == 0


def get_plone_site_by_args(args, requestor):
    if getattr(args, 'site', None):
        site = args.site
        setattr(args, 'picked_site', site)
        return site
    if getattr(args, 'pick_site', False):
        sites = get_sites(requestor)
        if len(sites) != 1:
            print('ERROR: --pick-site is ambiguous:')
            print('Expected exactly one site, found', len(sites))
            sys.exit(1)
        site = sites[0]['path']
        setattr(args, 'picked_site', site)
        return site
    if getattr(args, 'last_site', False):
        sites = get_sites(requestor)
        if len(sites) == 0:
            print('ERROR: No Plone site found.')
            sys.exit(1)
        site = sites[-1]['path']
        setattr(args, 'picked_site', site)
        return site
    if getattr(args, 'all_sites', None):
        sites = get_sites(requestor)
        if len(sites) == 0:
            print('ERROR: No Plone site found.')
            sys.exit(1)
        # Get and update site index stored in the arguments.  This is used for
        # iterating over the sites one by one.
        site_index = getattr(args, 'site_index', 0)
        if site_index >= len(sites):
            raise StopIteration
        site = sites[site_index]['path']
        setattr(args, 'site_index', site_index + 1)
        setattr(args, 'picked_site', site)
        return site
    return None


def get_sites(requestor):
    response = requestor.GET('list_plone_sites')
    return response.json()



from __future__ import print_function
from operator import itemgetter
from six.moves import filter
from six.moves import map
from six.moves import zip


class FakeTerminal(str):

    def __getattr__(self, name):
        if name == 'length':
            return len
        return self

    def __getattribute__(self, name):
        if name in ('center', 'ljust', 'lstrip', 'rjust', 'rstrip', 'strip'):
            return lambda text, *a, **kw: getattr(text, name)(*a, **kw)
        return str.__getattribute__(self, name)

    def __call__(self, text):
        return text


try:
    from blessed import Terminal
except ImportError:
    Terminal = FakeTerminal


TERMINAL = Terminal()
FLAGS = {
    'done': TERMINAL.green('done'),
    'orphan': TERMINAL.standout(TERMINAL.red('ORPHAN')),
    'outdated_fs_version': TERMINAL.standout(
        TERMINAL.red('FS-VERSION-OUTDATED')),
    'deferrable': TERMINAL.standout(TERMINAL.blue('DEFERRABLE')),
    'proposed': TERMINAL.blue('proposed')}


def print_table(data, titles=None, colspace=1):
    if titles:
        data.insert(0, list(map(TERMINAL.bright_black, titles)))

    column_lengths = list(map(max, zip(
        *[list(map(TERMINAL.length, row)) for row in data])))
    for row in data:
        for col_num, cell in enumerate(row):
            print(TERMINAL.ljust(cell, column_lengths[col_num] + colspace), end=' ')
        print('')


def upgrade_id_with_flags(upgrade, omit_flags=()):
    flags = set(FLAGS.keys()) - set(omit_flags)
    flagstext = ' '.join(FLAGS.get(flag) for flag in flags
                         if upgrade.get(flag, None))
    return ' '.join((colorize_api_id(upgrade['id']), flagstext)).strip()


def colorize_api_id(api_id):
    dest_version, profileid = api_id.split('@')
    return dest_version + '@' + colorize_profile_id(profileid)


def colorize_profile_id(profile_id):
    return TERMINAL.green(profile_id)


def colorized_profile_versions(profile):
    db_version = profile['db_version']
    fs_version = profile['fs_version']
    if db_version != fs_version:
        db_version = TERMINAL.red(db_version)
    return TERMINAL.bright_black(' / '.join((db_version, fs_version)))


def colorized_profile_flags(profile):
    flags = []
    if profile['outdated_fs_version']:
        flags.append(FLAGS['outdated_fs_version'])

    proposed = len(list(filter(itemgetter('proposed'), profile['upgrades'])))
    if proposed:
        flags.append(TERMINAL.blue('{0} proposed'.format(proposed)))

    orphans = len(list(filter(itemgetter('orphan'), profile['upgrades'])))
    if orphans:
        flags.append(TERMINAL.standout_red('{0} orphan'.format(orphans)))

    return ' '.join(flags)



from contextlib import contextmanager
from ftw.upgrade.gatherer import flatten_upgrades
from ftw.upgrade.interfaces import IUpgradeStepRecorder
from functools import partial
from operator import itemgetter
from Products.CMFCore.utils import getToolByName
from six.moves import map
from zope.component import getMultiAdapter

import os
import re


DISABLE_UPGRADE_STEP_MARKING_KEY = 'ftw_upgrade_disable_upgrade_step_marking'
ALL_FLAG = 'all'


@contextmanager
def no_upgrade_step_marking(*only_profiles):
    if only_profiles:
        value = ','.join(map(partial(re.sub, '^profile-', ''), only_profiles))
    else:
        value = ALL_FLAG

    previous_value = os.environ.get(DISABLE_UPGRADE_STEP_MARKING_KEY, None)
    os.environ[DISABLE_UPGRADE_STEP_MARKING_KEY] = value

    try:
        yield
    finally:
        if previous_value:
            os.environ[DISABLE_UPGRADE_STEP_MARKING_KEY] = previous_value
        else:
            os.environ.pop(DISABLE_UPGRADE_STEP_MARKING_KEY, None)


def profile_installed(event):
    if not event.profile_id:
        return

    if not event.full_import:
        return

    disabled_for_profiles = os.environ.get(DISABLE_UPGRADE_STEP_MARKING_KEY, '').split(',')
    profile = re.sub('^profile-', '', event.profile_id)
    if profile in disabled_for_profiles or ALL_FLAG in disabled_for_profiles:
        return

    portal = getToolByName(event.tool, 'portal_url').getPortalObject()
    recorder = getMultiAdapter((portal, profile), IUpgradeStepRecorder)

    list(map(recorder.mark_as_installed, map(
        itemgetter('sdest'), flatten_upgrades(
            event.tool.listUpgrades(profile, show_old=True)))))



from ftw.upgrade.command.terminal import FakeTerminal
from ftw.upgrade.command.terminal import TERMINAL
from unittest import TestCase


class TestFakeTerminal(TestCase):
    """The FakeTerminal is used instead of the blessed terminal when there are import
    problems, such as python not built with _curses support.
    The FakeTerminal should act similar to the Terminal class but not do things such
    as formatting / colorization.
    """

    def test_concatenating_colors(self):
        term = FakeTerminal()
        self.assertEqual(
            'foo',
            term.standout + term.green + term.red_bold + 'foo' + term.normal)

    def test_calling_colors(self):
        term = FakeTerminal()
        self.assertEqual(
            'foo',
            term.red_bold(term.standout('foo')))

    def test_length(self):
        self.assertEqual(3, FakeTerminal().length('foo'))

    def test_ljust(self):
        self.assertEqual(TERMINAL.ljust('foo', 10),
                         FakeTerminal().ljust('foo', 10))
        self.assertEqual(TERMINAL.ljust(u'foo', 10),
                         FakeTerminal().ljust(u'foo', 10))

    def test_rjust(self):
        self.assertEqual(TERMINAL.rjust('foo', 10),
                         FakeTerminal().rjust('foo', 10))

    def test_center(self):
        self.assertEqual(TERMINAL.center('foo', 10),
                         FakeTerminal().center('foo', 10))

    def test_strip(self):
        self.assertEqual(TERMINAL.strip('  foo  '),
                         FakeTerminal().strip('  foo  '))

    def test_rstrip(self):
        self.assertEqual(TERMINAL.rstrip('  foo  '),
                         FakeTerminal().rstrip('  foo  '))

    def test_lstrip(self):
        self.assertEqual(TERMINAL.lstrip('  foo  '),
                         FakeTerminal().lstrip('  foo  '))



from __future__ import print_function
from ftw.upgrade.command.terminal import TERMINAL
from ftw.upgrade.command.utils import find_egginfo
from ftw.upgrade.command.utils import find_package_namespace_path
from ftw.upgrade.directory.scaffold import UpgradeStepCreator
from path import Path

import argparse
import sys


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    The "create" command creates a new upgrade step in the "upgrades" \
directory of this package.

    The created upgrade step works only when the "upgrades" directory is \
registered using the "{t.bold_green}upgrade-step:directory{t.normal}" \
directive in ZCML.

    The command creates a new directory within the "upgrades" directory and \
adds an "upgrade.py" file with a default upgrade step. \
The "title" argument is used for naming the directory as well as the \
upgrade step class in the generated "upgrade.py" file.

{t.bold}EXAMPLES:{t.normal}
[quote]
$ bin/upgrade create AddIndexesToCatalog
$ bin/upgrade create AddIndexesToCatalog --path src/my/package/upgrades
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser('create',
                                  help='Create a new upgrade step.',
                                  description=DOCS)
    command.set_defaults(func=create_command)

    command.add_argument('--path', '-p',
                         metavar='upgrades-directory',
                         dest='upgrades_directory',
                         help='Path to the upgrades directory.'
                         ' The default path is searched by resolving the top'
                         ' level in the egg-info and looking for an "upgrades"'
                         ' directory.'
                         ' If no egg-info is found or there is no or multiple'
                         ' "upgrades" directories, the user has to provide the'
                         ' path explicitly.',
                         default=None,
                         type=upgrades_path)

    command.add_argument('title',
                         help='Title of the upgrade step.'
                         ' Either in camel case or with underscores.')


def create_command(args):
    upgrades_directory = args.upgrades_directory
    if upgrades_directory is None:
        upgrades_directory = default_upgrades_directory()

    if upgrades_directory is None:
        print('ERROR: Please provide the path to '
              'the upgrades directory with --path.', file=sys.stderr)
        sys.exit(1)

    creator = UpgradeStepCreator(upgrades_directory)
    upgrade_step_directory = creator.create(args.title)
    print('Created upgrade step at:', upgrade_step_directory)


def upgrades_path(path):
    path = Path(path).abspath()

    if not path.isdir():
        raise argparse.ArgumentTypeError(
            '"{0}" does not exist or is not a directory'.format(path))

    return path


def default_upgrades_directory():
    egginfo = find_egginfo()
    if not egginfo:
        return None

    package_namespace_path = find_package_namespace_path(egginfo)
    upgrades_dirs = tuple(package_namespace_path.walkdirs('upgrades'))
    if len(upgrades_dirs) == 0:
        print('WARNING: no "upgrades" directory could be found.', file=sys.stderr)
        return None

    if len(upgrades_dirs) > 1:
        print('WARNING: more than one "upgrades" directory found.', file=sys.stderr)
        return None

    return upgrades_dirs[0]



from ftw.upgrade.tests.base import CommandAndInstanceTestCase

import json


class TestSitesCommand(CommandAndInstanceTestCase):

    def setUp(self):
        super(TestSitesCommand, self).setUp()
        self.write_zconf_with_test_instance()

    def test_sites_help(self):
        self.upgrade_script('sites --help')

    def test_listing_sites_as_table(self):
        exitcode, output = self.upgrade_script('sites')
        self.assertEqual(0, exitcode)
        self.assertEqual('/plone               Plone site\n', output)

    def test_listing_sites_as_json(self):
        exitcode, output = self.upgrade_script('sites --json')
        self.assertEqual(0, exitcode)
        self.assert_json_equal([{'id': 'plone',
                                 'path': '/plone',
                                 'title': 'Plone site'}],
                               json.loads(output))

    def test_error_when_no_site_reachable(self):
        self.layer['root_path'].joinpath('parts/instance').rmtree()
        exitcode, output = self.upgrade_script('sites', assert_exitcode=False)
        self.assertEqual(1, exitcode)
        self.assertMultiLineEqual(
            'ERROR: No running Plone instance detected.\n',
            output)



from ftw.testbrowser import browsing
from ftw.upgrade.tests.base import JsonApiTestCase


class TestZopeAppJsonApi(JsonApiTestCase):

    def setUp(self):
        self.app = self.layer['app']

    @browsing
    def test_api_discovery(self, browser):
        self.api_request('GET', '', context=self.app)

        self.assert_json_equal(
            {'api_version': 'v1',
             'actions':
                 [{'name': 'current_user',
                   'required_params': [],
                   'description': 'Return the current user when authenticated properly.'
                   ' This can be used for testing authentication.',
                   'request_method': 'GET'},

                  {'name': 'list_plone_sites',
                   'required_params': [],
                   'description': 'Returns a list of Plone sites.',
                   'request_method': 'GET'}]},

            browser.json)

        self.assertTrue(browser.body.endswith(b'\n'),
                        'There should always be a trailing newline.')

    @browsing
    def test_list_plone_sites(self, browser):
        self.api_request('GET', 'list_plone_sites', context=self.app)

        self.assert_json_equal(
            [{'id': 'plone',
              'path': '/plone',
              'title': 'Plone site'}],
            browser.json)

    @browsing
    def test_current_user(self, browser):
        self.api_request('GET', 'current_user', context=self.app)
        self.assertEqual('admin', browser.json)

    @browsing
    def test_requiring_available_api_version_by_url(self, browser):
        self.api_request('GET', 'v1/list_plone_sites', context=self.app)
        self.assert_json_equal(
            [{'id': 'plone',
              'path': '/plone',
              'title': 'Plone site'}],
            browser.json)

    @browsing
    def test_requiring_wrong_api_version_by_url(self, browser):
        with self.expect_api_error(status=404,
                                   message='Wrong API version',
                                   details='The API version "v100" is not available.'):
            self.api_request('GET', 'v100/list_plone_sites', context=self.app)

        self.assertTrue(browser.body.endswith(b'\n'),
                        'There should always be a trailing newline.')

    @browsing
    def test_requesting_unkown_action(self, browser):
        with self.expect_api_error(status=404,
                                   message='Unkown API action',
                                   details='There is no API action "something".'):
            self.api_request('GET', 'something', context=self.app)



from ftw.upgrade.transactionnote import TransactionNote
from six.moves import range
from unittest import TestCase

import transaction


class TestTransactionNote(TestCase):

    def setUp(self):
        transaction.begin()

    def tearDown(self):
        self.assertLess(len(transaction.get().description), 65535,
                        'Transaction note should never be longer than 65535')
        transaction.abort()

    def test_transaction_note_is_updated(self):
        note = TransactionNote()
        note.add_upgrade('my.package:default', ('1','1'), 'Migrate objects')
        note.add_upgrade('my.package:default', ('1702',), 'Remove utility')
        note.set_transaction_note()

        self.assertEqual(
            u'my.package:default -> 1.1 (Migrate objects)\n'
            u'my.package:default -> 1702 (Remove utility)',
            transaction.get().description)

    def test_description_is_removed_when_note_gets_too_long(self):
        # Transaction note size is limited to 65535 characters
        description = 'A' * (65535 // 2)

        note = TransactionNote()
        note.add_upgrade('my.package:default', ('1000',), description)
        note.add_upgrade('my.package:default', ('1001',), description)
        note.set_transaction_note()

        # Prevent from printing the very long description in the assertion
        # message by not using assertIn..
        assert 'AAAA' not in transaction.get().description, \
            'Description seems not to be removed from too long' + \
            ' transaction note.'

        self.assertEqual(
            u'my.package:default -> 1000\n'
            u'my.package:default -> 1001',
            transaction.get().description)

    def test_cropped_when_too_long_even_without_description(self):
        profileid = 'my.package:default'

        transaction.get().note('Some notes..')

        note = TransactionNote()
        for destination in range(1, (65535 // len(profileid)) + 2):
            note.add_upgrade(profileid, (str(destination),), '')
        note.set_transaction_note()

        result = transaction.get().description
        expected_start = 'Some notes..\nmy.package:default -> 1\n'
        self.assertTrue(
            result.startswith(expected_start),
            ('Expected transaction note to start with "%s",'
             ' but it started with "%s"') % (
                expected_start, result[:50]))

        self.assertTrue(
            result.endswith('...'),
            'Expected transaction note to be cropped, ending with "..." '
            'but it ends with "%s"' % result[-30:])

    def test_cropped_according_to_already_existing_notes(self):
        # When the transaction note has already the maximum length
        # we cannot add anymore infos..
        profileid = 'my.package:default'
        transaction.get().note('.' * 65533)

        TransactionNote().add_upgrade(profileid, ('1000',), '')
        TransactionNote().set_transaction_note()

        self.assertLess(
            len(transaction.get().description), 65535,
            'Transaction note is too long, should be less than 65535')



from datetime import datetime
from ftw.builder import Builder
from ftw.upgrade import UpgradeStep
from ftw.upgrade.exceptions import CyclicDependencies
from ftw.upgrade.exceptions import UpgradeNotFound
from ftw.upgrade.gatherer import extend_auto_upgrades_with_human_formatted_date_version
from ftw.upgrade.gatherer import UpgradeInformationGatherer
from ftw.upgrade.interfaces import IUpgradeInformationGatherer
from ftw.upgrade.tests.base import UpgradeTestCase
from Products.CMFPlone.utils import getFSVersionTuple
from unittest import TestCase
from zope.component import queryAdapter
from zope.component import getMultiAdapter
from zope.interface.verify import verifyClass


class NotDeferrableUpgrade(UpgradeStep):
    """Test that attribute value is used, not attribtue presence."""

    deferrable = False

    def __call__(self):
        pass


class TestUpgradeInformationGatherer(UpgradeTestCase):

    def test_implements_interface(self):
        verifyClass(IUpgradeInformationGatherer, UpgradeInformationGatherer)

    def test_not_yet_installed_upgrades_are_marked_as_proposed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1001', to='1002')))

        with self.package_created():
            self.install_profile('the.package:default', '1001')
            self.assert_gathered_upgrades({
                    'the.package:default': {'proposed': ['1002'],
                                            'done': ['1001']}})

    def test_filtering_proposed_upgrades(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1000', to='1001'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1001', to='1002')))

        with self.package_created():
            self.install_profile('the.package:default', '1001')
            self.assert_gathered_upgrades(
                {'the.package:default': {'proposed': ['1002'],
                                         'done': []}},
                proposed_only=True)

    def test_profile_with_outdated_fs_version_is_flagged(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('outdated')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1111', '2222'))
                                  .with_fs_version('2222')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('2222', '3333')))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('correct')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1111', '2222'))
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('2222', '3333'))
                                  .with_fs_version('3333'))

        with self.package_created():
            self.install_profile('the.package:outdated', '2222')
            self.install_profile('the.package:correct', '2222')
            self.assert_outdated_profiles(['the.package:outdated'],
                                          ignore=[u'plone.app.jquerytools:default',
                                                  u'plone.app.jquery:default'])
            self.assert_gathered_upgrades({
                    'the.package:outdated': {'proposed': ['3333'],
                                            'done': ['2222'],
                                            'outdated_fs_version': ['3333']},
                    'the.package:correct': {'proposed': ['3333'],
                                            'done': ['2222'],
                                            'outdated_fs_version': []}})

    def test_profile_with_no_upgrades_is_not_listed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('without_upgrades')
                                  .with_fs_version('1'))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('with_upgrades')
                                  .with_fs_version('2')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1', to='2')))

        with self.package_created():
            self.install_profile('the.package:without_upgrades')
            self.install_profile('the.package:with_upgrades')
            self.assertIn('the.package:with_upgrades', self.get_listed_profiles(),
                          'Profiles with upgrades should be listed.')
            self.assertNotIn('the.package:without_upgrades', self.get_listed_profiles(),
                          'Profiles without upgrades should not be listed.')

    def test_profile_only_listed_when_installed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(Builder('plone upgrade step')
                                                .upgrading('1', to='2')))
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('uninstall'))

        with self.package_created():
            self.assertNotIn('the.package:default', self.get_listed_profiles(),
                             'Not yet installed profiles should not be listed.')

            self.install_profile('the.package:default')
            self.assertIn('the.package:default', self.get_listed_profiles(),
                          'Installed profiles should be listed.')

            if getFSVersionTuple() > (5, 1):
                installer = getMultiAdapter(
                    (self.portal, self.layer['request']), name='installer')
                installer.uninstall_product('the.package')
            else:
                self.portal_quickinstaller.uninstallProducts(['the.package'])
            self.assertNotIn('the.package:default', self.get_listed_profiles(),
                             'Packages uninstalled by quickinstaller should not be listed.')

    def test_plone_profile_is_removed(self):
        self.assertNotIn('Products.CMFPlone:plone',
                         self.get_listed_profiles(filter_package=None),
                         'Products.CMFPlone should never be listed.')

    def test_profiles_are_ordered_by_dependencies(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('one')
                                  .with_dependencies('the.package:two',
                                                     'the.package:four')
                                  .with_upgrade(self.default_upgrade()))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('two')
                                  .with_dependencies('the.package:three')
                                  .with_upgrade(self.default_upgrade()))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('three')
                                  .with_dependencies('the.package:four'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('four')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:one')
            self.install_profile('the.package:two')
            self.install_profile('the.package:three')
            self.install_profile('the.package:four')

            self.assertEqual(
                ['the.package:four',
                 'the.package:two',
                 'the.package:one'],
                self.get_listed_profiles())

    def test_cyclic_dependencies_raises_an_exception(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .named('foo')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:bar'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('bar')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:foo'))

        with self.package_created():
            self.install_profile('the.package:foo')
            self.install_profile('the.package:bar')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            with self.assertRaises(CyclicDependencies) as cm:
                gatherer.get_profiles()

            self.assertIn(('the.package:bar', 'the.package:foo'),
                          cm.exception.dependencies)

    def test_no_longer_existing_profiles_are_silently_removed(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(self.default_upgrade())
                                  .with_dependencies('the.package:removed'))

        self.package.with_profile(Builder('genericsetup profile')
                                  .named('removed')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:default')
            self.assertEqual(['the.package:removed', 'the.package:default'],
                             self.get_listed_profiles())

            from Products.GenericSetup import profile_registry
            profile_registry.unregisterProfile('removed', 'the.package')

            self.assertEqual(['the.package:default'],
                             self.get_listed_profiles())

    def test_not_installable_products_are_not_checked_if_uninstalled(self):
        # Some profiles are not associated with a product.
        # Since we want those profiles to appear in manage-upgrades, we dont verify in the
        # quickinstaller that the product is installed, since it is always not installed because
        # there is no known product.
        # - Products.CMFEditions:CMFEditions
        # - Products.TinyMCE:TinyMCE
        # - plone.app.discussion:default
        # - plone.formwidget.autocomplete:default (depends on Plone version)

        profiles = self.get_listed_profiles(filter_package=None)
        if getFSVersionTuple() < (5,):
            self.assertIn('Products.CMFEditions:CMFEditions', profiles)
            self.assertIn('Products.TinyMCE:TinyMCE', profiles)
        self.assertIn('plone.app.discussion:default', profiles)

    def test_profile_infos(self):
        self.package.with_profile(Builder('genericsetup profile')
                                  .with_upgrade(self.default_upgrade()))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')

            self.assertDictContainsSubset(
                {'product': 'the.package',
                 'id': u'the.package:default',
                 'db_version': u'1000',
                 'version': u'1001',
                 'outdated_fs_version': False,
                 'title': u'the.package',
                 'description': u'',
                 'type': 2,
                 },

                self.get_profiles_by_ids()['the.package:default'])

    def test_deferrable_profile_infos(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .calling(NotDeferrableUpgrade)))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')

            profiles = self.get_profiles_by_ids()
            profile_info = profiles['the.package:default']
            upgrade_info = profile_info['upgrades'][0]

            self.assertDictContainsSubset(
                {'profile': 'the.package:default',
                 'done': False,
                 'proposed': True,
                 'orphan': False,
                 'deferrable': False,
                 'outdated_fs_version': False,
                 },
                upgrade_info)

    def test_upgrade_infos(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step')
                          .upgrading('1000', to='1001')
                          .titled(u'Add action')
                          .with_description('Some details...')))

        with self.package_created():
            self.install_profile('the.package:default', version='1000')
            profile_info = self.get_profiles_by_ids()['the.package:default']
            upgrade_info = profile_info['upgrades'][0]

            self.assertDictContainsSubset(
                {'title': u'Add action',
                 'description': u'Some details...',
                 'profile': 'the.package:default',
                 'api_id': '1001@the.package:default',
                 'ssource': '1000',
                 'source': ('1000',),
                 'sdest': '1001',
                 'dest': ('1001',),
                 'done': False,
                 'proposed': True,
                 'orphan': False,
                 'deferrable': False,
                 'outdated_fs_version': False,
                 'haspath': ('1001',)},
                upgrade_info)

    def test_upgrade_infos_proposes_deferrable_by_default(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')
            self.assert_gathered_upgrades({
                    'the.package:default': {
                        'done': [],
                        'proposed': ['20110101000000'],
                        'orphan': []}})

            profiles = self.get_profiles_by_ids()
            profile_info = profiles['the.package:default']
            upgrade_info = profile_info['upgrades'][0]

            self.assertDictContainsSubset(
                {'profile': 'the.package:default',
                 'done': False,
                 'proposed': True,
                 'orphan': False,
                 'deferrable': True,
                 'outdated_fs_version': False,
                 },
                upgrade_info)

    def test_deferrable_upgrades_are_proposed_when_specified(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 1, 1))
                          .as_deferrable()))

        with self.package_created():
            self.install_profile('the.package:default', version='1')
            self.clear_recorded_upgrades('the.package:default')
            self.assert_gathered_upgrades({
                    'the.package:default': {
                        'done': [],
                        'proposed': ['20110101000000'],
                        'orphan': []}},
                    propose_deferrable=True)

            profiles = self.get_profiles_by_ids(propose_deferrable=True)
            profile_info = profiles['the.package:default']
            upgrade_info = profile_info['upgrades'][0]

            self.assertDictContainsSubset(
                {'profile': 'the.package:default',
                 'done': False,
                 'proposed': True,
                 'orphan': False,
                 'deferrable': True,
                 'outdated_fs_version': False,
                 },
                upgrade_info)

    def test_orphane_upgrades_are_marked(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2013, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='20120101000000')
            self.record_installed_upgrades('the.package:default', '20120101000000')
            self.assert_gathered_upgrades({
                    'the.package:default': {
                        'done': ['1001', '20120101000000'],
                        'proposed': ['20110101000000', '20130101000000'],
                        'orphan': ['20110101000000']}})

    def test_do_not_propose_future_upgrades_marked_as_executed(self):
        """When the installed version is older than an upgrade step, but the upgrade
        step was already marked as executed, the upgrade step should not be proposed.

        This is mainly useful for when the IUpgradeStepRecorder is extend with additional
        logic.
        """
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('plone upgrade step').upgrading('1000', to='1001'))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2013, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default', version='20110101000000')
            self.record_installed_upgrades(
                'the.package:default', '20110101000000', '20130101000000')
            self.assert_gathered_upgrades({
                    'the.package:default': {
                        'done': ['1001', '20110101000000', '20130101000000'],
                        'proposed': ['20120101000000'],
                        'orphan': []}})

    def test_human_readable_timestamp_versions_are_added_to_profiles(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step')
                          .to(datetime(2011, 12, 30, 17, 45, 59))))

        with self.package_created():
            self.install_profile('the.package:default')

            self.assertDictContainsSubset(
                {'id': u'the.package:default',
                 'db_version': u'20111230174559',
                 'formatted_db_version': u'2011/12/30 17:45',
                 'version': u'20111230174559',
                 'formatted_version': u'2011/12/30 17:45'},
                self.get_profiles_by_ids()['the.package:default'])

    def test_human_readable_timestamp_versions_are_added_to_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to('20141117093040'))
            .with_upgrade(Builder('ftw upgrade step').to('20141230104550')))

        with self.package_created():
            self.install_profile('the.package:default')
            profile_info = self.get_profiles_by_ids()['the.package:default']
            upgrade_info = profile_info['upgrades'][1]

            self.assertDictContainsSubset(
                {'ssource': '20141117093040',
                 'fsource': '2014/11/17 09:30',
                 'sdest': '20141230104550',
                 'fdest': '2014/12/30 10:45'},
                upgrade_info)

    def test_get_upgrades_by_api_ids(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1))))

        with self.package_created():
            self.install_profile('the.package:default')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            upgrade_info, = gatherer.get_upgrades_by_api_ids(
                '20110101000000@the.package:default')
            self.assertDictContainsSubset(
                {'api_id': u'20110101000000@the.package:default',
                 'sdest': u'20110101000000'},
                upgrade_info)

    def test_get_upgrades_by_api_ids_orders_upgrades(self):
        self.package.with_profile(
            Builder('genericsetup profile')
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2011, 1, 1)))
            .with_upgrade(Builder('ftw upgrade step').to(datetime(2012, 2, 2))))

        with self.package_created():
            self.install_profile('the.package:default')

            gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
            self.assertEqual(
                gatherer.get_upgrades_by_api_ids('20110101000000@the.package:default',
                                                 '20120202000000@the.package:default'),
                gatherer.get_upgrades_by_api_ids('20120202000000@the.package:default',
                                                 '20110101000000@the.package:default'))

    def test_get_upgrades_by_api_ids_raises_upgrade_not_found(self):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        with self.assertRaises(UpgradeNotFound) as cm:
            gatherer.get_upgrades_by_api_ids('foo@bar:default')
        self.assertEqual('foo@bar:default', cm.exception.api_id)
        self.assertEqual('The upgrade "foo@bar:default" could not be found.',
                         str(cm.exception))

    def get_listed_profiles(self, filter_package='the.package'):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles()
        profiles = [profile['id'] for profile in result]
        if filter_package:
            profiles = [
                profile for profile in profiles
                if profile.startswith(filter_package)
            ]
        return profiles

    def get_profiles_by_ids(self, **kwargs):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles(**kwargs)
        return dict([(profile['id'], profile) for profile in result])

    def assert_outdated_profiles(self, expected_profiles, ignore=()):
        gatherer = queryAdapter(self.portal_setup, IUpgradeInformationGatherer)
        result = gatherer.get_profiles()
        got_profiles = [profile['id'] for profile in result
                        if profile['outdated_fs_version'] and profile['id'] not in ignore]
        self.assertEqual(expected_profiles, got_profiles,
                         'Unexpected outdated fs versions for profiles.')


class TestExtendAutoUpgradesWithHumanFormattedDateVersion(TestCase):

    def test_formats_timestamp_source_versions(self):
        input = {'ssource': '20141214183059',
                 'sdest': '1'}
        expected = {'ssource': '20141214183059',
                    'fsource': '2014/12/14 18:30',
                    'sdest': '1'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_formats_timestamp_dest_versions(self):
        input = {'ssource': '1',
                 'sdest': '20141214183059'}
        expected = {'ssource': '1',
                    'sdest': '20141214183059',
                    'fdest': '2014/12/14 18:30'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_does_not_format_non_timestamp_versions(self):
        input = {'ssource': '1',
                 'sdest': '2'}
        expected = {'ssource': '1',
                    'sdest': '2'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def test_does_not_fail_on_non_timestamp_versions_looking_like_timestamps(self):
        input = {'ssource': '10000000000000',
                 'sdest': '1'}
        expected = {'ssource': '10000000000000',
                    'sdest': '1'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

        input = {'ssource': '1',
                 'sdest': '10000000000000'}
        expected = {'ssource': '1',
                    'sdest': '10000000000000'}
        self.assertDictEqual(expected, self.format_upgrade_step(input))

    def format_upgrade_step(self, upgrade_step):
        profiles = [{'upgrades': [upgrade_step]}]
        output = extend_auto_upgrades_with_human_formatted_date_version(profiles)
        return output[0]['upgrades'][0]



from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility
from zope.component.hooks import getSite


def recook_resources():
    for name in ('portal_css', 'portal_javascripts'):
        registry = getToolByName(getSite(), name, None)
        if registry is not None:
            registry.cookResources()

    # Plone 5: clear all bundles
    registry = getUtility(IRegistry)
    for key in [key for key in registry.records.keys()
                if (key.startswith('plone.bundles/')
                    and key.endswith('.last_compilation'))]:
        registry[key] = None



from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from DateTime import DateTime
from ftw.upgrade.helpers import update_security_for
from functools import partial
from operator import methodcaller
from persistent.mapping import PersistentMapping
from plone.app.relationfield.event import extract_relations
from plone.app.textfield import IRichText
from plone.app.textfield import IRichTextValue
from plone.app.uuid.utils import uuidToObject
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.utils import createContent
from plone.dexterity.utils import iterSchemata
from plone.namedfile.interfaces import INamedBlobFileField
from plone.namedfile.interfaces import INamedBlobImageField
from plone.namedfile.interfaces import INamedField
from plone.uuid.interfaces import IMutableUUID
from plone.uuid.interfaces import IUUID
from Products.CMFPlone.interfaces import constrains
from Products.CMFPlone.utils import getFSVersionTuple
from six.moves import filter
from six.moves import map
from z3c.relationfield.event import _setRelation
from z3c.relationfield.interfaces import IRelation
from z3c.relationfield.relation import create_relation
from zope.annotation import IAnnotations
from zope.component import getUtility
from zope.container.contained import notifyContainerModified
from zope.event import notify
from zope.interface import Interface
from zope.intid.interfaces import IIntIds
from zope.keyreference.interfaces import IKeyReference
from zope.lifecycleevent import ObjectModifiedEvent
from zope.schema import getFieldsInOrder

import logging
import pkg_resources
import six


try:
    pkg_resources.get_distribution('Products.Archetypes')
except pkg_resources.DistributionNotFound:
    class IBaseObject(Interface):
        pass

    class IComputedField(Interface):
        pass
else:
    from Products.Archetypes.interfaces import IBaseObject
    from Products.Archetypes.interfaces import IComputedField

try:
    pkg_resources.get_distribution('plone.app.blob')
except pkg_resources.DistributionNotFound:
    class IBlobWrapper(Interface):
        pass
else:
    from plone.app.blob.interfaces import IBlobWrapper

try:
    pkg_resources.get_distribution('archetypes.referencebrowserwidget')
except pkg_resources.DistributionNotFound:
    class IATReferenceField(Interface):
        pass
else:
    from archetypes.referencebrowserwidget.interfaces import IATReferenceField


DISABLE_FIELD_AUTOMAPPING = 1
IGNORE_UNMAPPED_FIELDS = 2
BACKUP_AND_IGNORE_UNMAPPED_FIELDS = 4
IGNORE_STANDARD_FIELD_MAPPING = 8
IGNORE_DEFAULT_IGNORE_FIELDS = 16
SKIP_MODIFIED_EVENT = 32

UNMAPPED_FIELDS_BACKUP_ANN_KEY = 'ftw.upgrade.migration:fields_backup'

LOG = logging.getLogger('ftw.upgrade.migration')


DEFAULT_ATTRIBUTES_TO_COPY = (
    '__ac_local_roles__',
    '__ac_local_roles_block__',
    '__annotations__',
    '_count',
    '_mt_index',
    '_owner',
    '_tree',
    'workflow_history',
)


DUBLIN_CORE_IGNORES = (
    'allowDiscussion',
    'contributors',
    'creators',
    'nextPreviousEnabled',
    'rights',
    'language',
    'relatedItems',
)


DEFAULT_IGNORED_FIELDS = (
    # ID, creation date and modification date are set by method, not by field.
    'id',
    'creation_date',
    'modification_date',

    # Constrain types are not regular behaviors.
    'constrainTypesMode',
    'locallyAllowedTypes',
    'immediatelyAddableTypes',

    # The location field is no longer available in Dexterity.
    'location',

    # The presentation field is no longer available in Dexterity.
    'presentation',
)


DEFAULT_OMIT_INDEXES = (
    'SearchableText',  # expensive; should be calculated afterwards
)


STANDARD_FIELD_MAPPING = {
    'subject': 'subjects',
    'allowDiscussion': 'allow_discussion',
    'effectiveDate': 'effective',
    'expirationDate': 'expires',
    'excludeFromNav': 'exclude_from_nav',
    'tableContents': 'table_of_contents',
}


class IgnoreField(Exception):
    """Exception for indicating that the current field should
    be silently ignored.
    Used internally only.
    """


class FieldNotMappedError(ValueError):
    """A field was not mapped properly.
    """


class FieldsNotMappedError(ValueError):
    """A field was not mapped properly.
    """

    message_template = (
        'Some fields are not mapped correctly when migrating'
        ' from "{old_type}" to "{new_type}":\n\n'
        'Not mapped:\n- {not_mapped}\n\n'
        'Target fields:\n- {target_fields}\n\n'
    )

    def __init__(self, not_mapped, old_type, new_type, target_fields):
        if getFSVersionTuple() > (5, ):
            raise NotImplementedError(
                'The inplace migrator migrates from Archetypes.'
                ' Plone 5 has no Archetypes objects.')
        super(FieldsNotMappedError, self).__init__(
            self.message_template.format(
                not_mapped='\n- '.join(sorted(not_mapped)),
                old_type=old_type,
                new_type=new_type,
                target_fields='\n- '.join(sorted(target_fields))))
        self.not_mapped_fields = not_mapped
        self.old_type = old_type
        self.new_type = new_type
        self.target_fields = target_fields


class InplaceMigrator(object):
    """The inplace migrator allows to easily migrate object inplace
    to dexterity objects.

    It supports Archetypes and Dexterity frameworks as source but only can
    produce Dexterity objects as destination.
    """

    def __init__(self,
                 new_portal_type,
                 field_mapping=None,
                 options=0,
                 ignore_fields=(),
                 attributes_to_migrate=DEFAULT_ATTRIBUTES_TO_COPY,
                 additional_steps=(),
                 omit_indexes=DEFAULT_OMIT_INDEXES):

        self.new_portal_type = new_portal_type
        self.field_mapping = field_mapping or {}
        self.options = options
        self.attributes_to_migrate = attributes_to_migrate
        self.omit_indexes = omit_indexes

        self.ignore_fields = list(ignore_fields)
        if not (options & IGNORE_DEFAULT_IGNORE_FIELDS):
            self.ignore_fields.extend(DEFAULT_IGNORED_FIELDS)

        self.steps_before_clone = (
            self.dump_and_remove_references,
        )
        self.steps_after_clone = (
            self.migrate_intid,
            self.migrate_field_values,
            self.add_relations_to_relation_catalog,
            self.migrate_properties,
            self.migrate_constrain_types_configuration,
            self.update_creators,
        )
        self.additional_steps = additional_steps
        self.final_steps = (
            self.update_creation_date,
            self.update_security,
            self.trigger_modified_event,
            self.update_modification_date,
        )

    def migrate_object(self, old_object):
        list(map(lambda func: func(old_object),
                 self.steps_before_clone))
        new_object = self.clone_object(old_object)
        list(map(lambda func: func(old_object, new_object),
                 list(self.steps_after_clone)
                 + list(self.additional_steps)
                 + list(self.final_steps)))
        return new_object

    def dump_and_remove_references(self, old_object):
        """We can only remove the relations from the reference_catalog
        as long as we did not replace the object (clone_object()),
        because otherwise the catalog cannot lookup the object anymore.
        We need to remove references from the reference_catalog,
        because otherwise the references will stay forever.
        Usually, when having DX=>DX relations at the end, we should no
        longer have references in the (AT) reference_catalog but only
        in the zc.catalog (intid).
        """
        self.removed_field_values = {}

        if not IBaseObject.providedBy(old_object):
            return  # not AT

        for field in old_object.Schema().values():
            if not IATReferenceField.providedBy(field):
                continue

            value = field.getRaw(old_object)
            if not value:
                continue

            self.removed_field_values[field.__name__] = value
            field.set(old_object, ())  # removes references.

    def clone_object(self, old_object):
        new_object = self.construct_clone_for(old_object)
        self.migrate_id_and_uuid(old_object, new_object)
        self.migrate_attributes(old_object, new_object)
        new_object = self.replace_in_parent(old_object, new_object)
        return new_object

    def construct_clone_for(self, old_object):
        return createContent(self.new_portal_type)

    def migrate_id_and_uuid(self, old_object, new_object):
        new_object.id = old_object.id
        IMutableUUID(new_object).set(IUUID(old_object))

    def migrate_intid(self, old_object, new_object):
        if '_intids' not in dir(self):
            self._intids = getUtility(IIntIds)

        old_key = IKeyReference(old_object)
        new_key = IKeyReference(new_object)
        try:
            uid = self._intids.ids[old_key]
        except KeyError:
            # Key was missing in catalog
            uid = self._intids.register(new_object)
            LOG.info('Add previously unregistered object to IIntIds catalog %s', new_object)
        else:
            del self._intids.ids[old_key]
        finally:
            self._intids.refs[uid] = new_key
            self._intids.ids[new_key] = uid


    def migrate_attributes(self, old_object, new_object):
        old_object = aq_base(old_object)
        new_object = aq_base(new_object)

        for name in self.attributes_to_migrate:
            if hasattr(old_object, name):
                setattr(new_object, name, getattr(old_object, name))

    def replace_in_parent(self, old_object, new_object):
        parent = aq_parent(aq_inner(old_object))
        new_object = aq_base(new_object)
        position_in_parent = parent.getObjectPosition(old_object.id)

        parent._delOb(old_object.id)
        objects = [info for info in parent._objects
                   if info['id'] != new_object.id]
        objects = tuple(objects)
        objects += ({'id': new_object.id,
                     'meta_type': getattr(new_object, 'meta_type', None)},)
        parent._objects = objects
        parent._setOb(new_object.id, new_object)
        notifyContainerModified(parent)

        if hasattr(aq_base(parent), '_tree'):
            del parent._tree[new_object.id]
            parent._tree[new_object.id] = new_object

        parent.moveObjectToPosition(new_object.id, position_in_parent)

        return parent._getOb(new_object.id)

    def migrate_field_values(self, old_object, new_object):
        not_mapped = {}
        new_field_map = self.build_new_field_map(new_object)

        for old_fieldname, value in self.get_field_values(old_object):
            try:
                new_field = self.get_new_field(old_object,
                                               new_object,
                                               old_fieldname,
                                               new_field_map)

            except FieldNotMappedError:
                not_mapped[old_fieldname] = value
                continue

            except IgnoreField:
                continue

            else:
                self.set_field_value(new_object, new_field, value)

        if not_mapped and self.options & BACKUP_AND_IGNORE_UNMAPPED_FIELDS:
            annotations = IAnnotations(new_object)
            if UNMAPPED_FIELDS_BACKUP_ANN_KEY not in annotations:
                annotations[UNMAPPED_FIELDS_BACKUP_ANN_KEY] = (
                    PersistentMapping())

            annotations[UNMAPPED_FIELDS_BACKUP_ANN_KEY].update(not_mapped)

        elif not_mapped:
            raise FieldsNotMappedError(list(not_mapped.keys()),
                                       old_object.portal_type,
                                       new_object.portal_type,
                                       new_field_map)

    def get_new_field(self, old_object, new_object,
                      old_fieldname, new_field_map):
        if old_fieldname in self.field_mapping:
            new_fieldname = self.field_mapping[old_fieldname]
            if new_fieldname in new_field_map:
                return new_field_map[new_fieldname]

        if not (self.options & IGNORE_STANDARD_FIELD_MAPPING):
            if old_fieldname in STANDARD_FIELD_MAPPING:
                new_fieldname = STANDARD_FIELD_MAPPING[old_fieldname]
                if new_fieldname in new_field_map:
                    return new_field_map[new_fieldname]

        if not (self.options & DISABLE_FIELD_AUTOMAPPING):
            if old_fieldname in new_field_map:
                return new_field_map[old_fieldname]

        if self.options & IGNORE_UNMAPPED_FIELDS:
            raise IgnoreField()

        raise FieldNotMappedError()

    def get_field_values(self, old_object):
        if IBaseObject.providedBy(old_object):
            return self.get_at_field_values(old_object)
        elif IDexterityContent.providedBy(old_object):
            return self.get_dx_field_values(old_object)
        else:
            raise NotImplementedError('Only AT and DX is supported.')

    def get_at_field_values(self, old_object):
        for field in old_object.Schema().values():
            if IComputedField.providedBy(field):
                continue

            fieldname = field.__name__
            if fieldname in self.ignore_fields:
                continue

            value = self.removed_field_values.get(
                fieldname, field.getRaw(old_object))
            value = self.normalize_at_field_value(field, fieldname, value)
            yield fieldname, value

    def normalize_at_field_value(self, old_field, old_fieldname, value):
        recurse = partial(self.normalize_at_field_value,
                          old_field, old_fieldname)

        if isinstance(value, str):
            return recurse(value.decode('utf-8'))

        if isinstance(value, list):
            return list(map(recurse, value))

        if isinstance(value, tuple):
            return tuple(map(recurse, value))

        if isinstance(value, DateTime):
            return recurse(value.asdatetime().replace(tzinfo=None))

        return value

    def get_dx_field_values(self, old_object):
        no_value_marker = object()

        for schemata in iterSchemata(old_object):
            storage = schemata(old_object)

            for fieldname, field in getFieldsInOrder(schemata):
                if fieldname in self.ignore_fields:
                    continue

                value = getattr(storage, fieldname, no_value_marker)
                if value == no_value_marker:
                    continue

                value = self.normalize_dx_field_value(field, fieldname, value)
                yield fieldname, value

    def normalize_dx_field_value(self, old_field, old_fieldname, value):
        return value

    def set_field_value(self, new_object, field, value):
        value = self.prepare_field_value(new_object, field, value)
        field.set(field.interface(new_object), value)

    def prepare_field_value(self, new_object, field, value):
        recurse = partial(self.prepare_field_value, new_object, field)

        if isinstance(value, str):
            return recurse(value.decode('utf-8'))

        if isinstance(value, list):
            return list(map(recurse, value))

        if isinstance(value, tuple):
            return tuple(map(recurse, value))

        relation_fields = list(filter(
            IRelation.providedBy, (field, getattr(field, 'value_type', None))))
        if relation_fields and isinstance(value, six.text_type):
            target = uuidToObject(value)
            return create_relation('/'.join(target.getPhysicalPath()))

        if IRichText.providedBy(field) \
           and not IRichTextValue.providedBy(value):
            if not value:
                return None
            else:
                return recurse(field.fromUnicode(value))

        if INamedField.providedBy(field) and value is not None \
           and not isinstance(value, field._type):

            if value == '':
                return None

            if hasattr(value, 'get_size') and value.get_size() == 0:
                return None

            source_is_blobby = IBlobWrapper.providedBy(value)
            target_is_blobby = INamedBlobFileField.providedBy(field) or \
                               INamedBlobImageField.providedBy(field)

            if source_is_blobby and target_is_blobby:
                filename = value.filename
                if isinstance(filename, str):
                    filename = filename.decode('utf-8')

                new_value = field._type(
                    data='',  # empty blob, will be replaced
                    contentType=value.content_type,
                    filename=filename)
                if not hasattr(new_value, '_blob'):
                    raise ValueError(
                        ('Unsupported file value type {!r}'
                         ', missing _blob.').format(
                             new_value.__class__))

                # Simply copy the persistent blob object (with the file system
                # pointer) to the new value so that the file is not copied.
                # We assume that the old object is trashed and can therefore
                # adopt the blob file.
                new_value._blob = value.getBlob()
                return recurse(new_value)

            else:
                filename = value.filename
                if isinstance(filename, str):
                    filename = filename.decode('utf-8')

                data = value.data
                data = getattr(data, 'data', data)  # extract Pdata
                return recurse(field._type(
                    data=data,
                    contentType=value.content_type,
                    filename=filename))

        return value

    def build_new_field_map(self, new_object):
        fieldmap = {}

        for schemata in iterSchemata(new_object):
            for new_fieldname, field in getFieldsInOrder(schemata):
                if field.readonly:
                    continue

                fieldmap[new_fieldname] = field
                fieldmap['.'.join((field.interface.__identifier__,
                                   new_fieldname))] = field

        return fieldmap

    def add_relations_to_relation_catalog(self, old_object, new_object):
        for behavior_interface, name, relation in extract_relations(
                new_object):
            if isinstance(relation, (str, six.text_type)):
                # We probably got a UID, but we are working with intids
                # and can not do anything with it, so we skip it.
                LOG.warning('Got a invalid relation ({!r}), which is not '
                            'z3c.relationfield compatible.'.format(relation))
                continue

            _setRelation(new_object, name, relation)

    def migrate_properties(self, old_object, new_object):
        for item in old_object.propertyMap():
            key = item['id']
            if key == 'title':
                continue

            value = old_object.getProperty(key)
            if new_object.hasProperty(key):
                new_object._updateProperty(key, value)
            else:
                new_object._setProperty(key, value, item['type'])

    def migrate_constrain_types_configuration(self, old_object, new_object):
        old_ct = constrains.IConstrainTypes(old_object, None)
        if not old_ct:
            # No constrain types support on old object.
            # It might not be folderish.
            return

        old_mode = old_ct.getConstrainTypesMode()
        if old_mode == constrains.DISABLED:
            return

        new_ct = constrains.ISelectableConstrainTypes(new_object, None)
        if new_ct is None:
            return

        new_ct.setConstrainTypesMode(old_mode)

        if old_mode != constrains.ENABLED:
            return

        allowed_types = list(map(methodcaller('getId'),
                                 new_ct.getDefaultAddableTypes()))
        isallowed = allowed_types.__contains__
        new_ct.setLocallyAllowedTypes(
            list(filter(isallowed, old_ct.getLocallyAllowedTypes())))
        new_ct.setImmediatelyAddableTypes(
            list(filter(isallowed, old_ct.getImmediatelyAddableTypes())))

    def update_creators(self, old_object, new_object):
        """When the dublin core behavior is active, the creators are migrated already.
        But when the dublin core behavior is missing, we need to fix the creators list here.
        """
        if old_object.listCreators() != new_object.listCreators():
            new_object.setCreators(old_object.listCreators())

    def update_creation_date(self, old_object, new_object):
        new_object.creation_date = (
            old_object.created().asdatetime().replace(tzinfo=None))

    def update_security(self, old_object, new_object):
        update_security_for(new_object, reindex_security=False)

    def trigger_modified_event(self, old_object, new_object):
        if self.options & SKIP_MODIFIED_EVENT:
            return

        notify(ObjectModifiedEvent(new_object))

    def update_modification_date(self, old_object, new_object):
        new_object.setModificationDate(
            old_object.modified().asdatetime().replace(tzinfo=None))

        if not (self.options & SKIP_MODIFIED_EVENT):
            new_object.reindexObject(idxs=['modified'])



from ftw.upgrade.command.jsonapi import add_json_argument
from ftw.upgrade.command.jsonapi import add_requestor_authentication_argument
from ftw.upgrade.command.jsonapi import add_site_path_argument
from ftw.upgrade.command.jsonapi import error_handling
from ftw.upgrade.command.jsonapi import with_api_requestor
from ftw.upgrade.command.terminal import colorize_profile_id
from ftw.upgrade.command.terminal import colorized_profile_flags
from ftw.upgrade.command.terminal import colorized_profile_versions
from ftw.upgrade.command.terminal import print_table
from ftw.upgrade.command.terminal import TERMINAL
from ftw.upgrade.command.terminal import upgrade_id_with_flags


DOCS = """
{t.bold}DESCRIPTION:{t.normal}
    List profiles or upgrades of a Plone site.

{t.bold}LIST PROFILES:{t.normal}
    Listing profiles lists all profiles which are installed on this Plone \
site and have at least one upgrade step. The profiles are listed with profile \
ID, amount of proposed and orphan upgrades for this profile, the title as \
well as database and filesystem versions.

[quote]
    $ ./bin/upgrade list --site Plone --profiles
    $ ./bin/upgrade list --site Plone --profiles --auth admin:admin
    $ ./bin/upgrade list --site Plone --profiles --auth admin:admin --json
[/quote]

{t.bold}LIST PROPOSED UPGRADES:{t.normal}
    Listing proposed upgrades lists all upgrades which are proposed for this \
Plone site. Only profiles installed on this Plone site are respected.

[quote]
    $ ./bin/upgrade list --site Plone --upgrades
    $ ./bin/upgrade list --site Plone --upgrades --auth admin:admin
    $ ./bin/upgrade list --site Plone --upgrades --auth admin:admin --json
[/quote]
""".format(t=TERMINAL).strip()


def setup_argparser(commands):
    command = commands.add_parser(
        'list',
        help='List upgrades or profiles.',
        description=DOCS)
    command.set_defaults(func=list_command)
    add_requestor_authentication_argument(command)
    add_site_path_argument(command)
    add_json_argument(command)

    group = command.add_mutually_exclusive_group(required=True)
    group.add_argument('--upgrades', '-u',
                       help='List all proposed upgrades.',
                       dest='action',
                       action='store_const',
                       const='list_proposed_upgrades')

    group.add_argument('--profiles', '-p',
                       help='List all installed profiles.',
                       dest='action',
                       action='store_const',
                       const='list_profiles')


@with_api_requestor
@error_handling
def list_command(args, requestor):
    response = requestor.GET(args.action)
    if args.json:
        print(response.text)
        return

    if args.action == 'list_proposed_upgrades':
        return format_proposed_upgrades(response)
    else:
        return format_profiles(response)


def format_proposed_upgrades(response):
    proposed = []
    for upgrade in response.json():
        is_deferrable = upgrade.get('deferrable', False)

        omit_flags = ('proposed', 'orphan') if is_deferrable else ('proposed',)

        table_row = [upgrade_id_with_flags(upgrade, omit_flags=omit_flags),
                     TERMINAL.bold(upgrade.get('title')),
                     ]
        proposed.append(table_row)

    print(TERMINAL.bold('Proposed upgrades:'))
    print_table(proposed, ['ID:', 'Title:'])


def format_profiles(response):
    tabledata = []
    for profile in response.json():
        tabledata.append(
            [colorize_profile_id(profile['id']),
             colorized_profile_flags(profile),
             TERMINAL.bold(profile['title']),
             colorized_profile_versions(profile)])

    print(TERMINAL.bold('Installed profiles:'))
    print_table(tabledata, ['ID:', '', 'Title:', 'Versions (DB/FS):'])
